"""This module contains dataclasses used in the ``Scheduling`` environment."""
from __future__ import annotations

from dataclasses import dataclass

import numpy as np
from numpy.typing import NDArray

from qgym.custom_types import Gate
from qgym.envs.scheduling.rulebook import CommutationRulebook
from qgym.utils.gate_encoder import GateEncoder
from qgym.utils.random_circuit_generator import RandomCircuitGenerator


@dataclass
class SchedulingUtils:
    """Utils used in the ``Scheduling`` environment."""

    random_circuit_generator: RandomCircuitGenerator
    random_circuit_mode: str
    rulebook: CommutationRulebook
    gate_encoder: GateEncoder


@dataclass
class GateInfo:
    """Info of a specific gate used in the ``Scheduling`` environment."""

    cycle_length: int
    not_in_same_cycle: set[int]
    exclude: int = 0
    exclude_next_cycle: bool = False

    def reset(self) -> GateInfo:
        """Reset the object.

        To be used in the reset function of the ``Scheduling`` environment.

        Returns:
            Self.
        """
        self.exclude = 0
        self.exclude_next_cycle = False
        return self


@dataclass
class CircuitInfo:
    """Info of the circuit of the current episode of ``Scheduling`` environment."""

    encoded: list[Gate]
    names: NDArray[np.int_]
    acts_on: NDArray[np.int_]
    legal: NDArray[np.int8]
    dependencies: NDArray[np.int_]
    schedule: NDArray[np.int_]
    blocking_matrix: NDArray[np.int_]

    def reset(self, circuit: list[Gate] | None, utils: SchedulingUtils) -> CircuitInfo:
        """Reset the object.

        To be used in the reset function of the ``Scheduling`` environment.

        Returns:
            Self.
        """
        if circuit is None:
            circuit = utils.random_circuit_generator.generate_circuit(
                mode=utils.random_circuit_mode
            )

        self.blocking_matrix = utils.rulebook.make_blocking_matrix(circuit)
        self.encoded = utils.gate_encoder.encode_gates(circuit)
        self.schedule = np.full(len(circuit), -1, dtype=int)
        return self



"""This module contains some vanilla Rewarders for the ``Scheduling`` environment.

Usage:
    The rewarders in this module can be customized by initializing the rewarders with
    different values.

    .. code-block:: python

        from qgym.envs.scheduling import BasicRewarder

        rewarder = BasicRewarder(
            illegal_action_penalty = -1,
            update_cycle_penalty = -2,
            schedule_gate_bonus: = 3,
            )

    After initialization, the rewarders can be given to the ``Scheduling`` environment.

.. note::
    When implementing custom rewarders, they should inherit from ``qgym.Rewarder``.
    Furthermore, they must implement the ``compute_reward`` method. Which takes as input
    the old state, the new state and the given action. See the documentation of the
    ``qgym.envs.scheduling.scheduling`` module for more information on the state and
    action space.

"""
from __future__ import annotations

import numpy as np
from numpy.typing import NDArray

from qgym.envs.scheduling.scheduling_state import SchedulingState
from qgym.templates import Rewarder
from qgym.utils.input_validation import check_real, warn_if_negative, warn_if_positive


class BasicRewarder(Rewarder):
    """Basic rewarder for the ``Scheduling`` environment."""

    def __init__(
        self,
        illegal_action_penalty: float = -5.0,
        update_cycle_penalty: float = -1.0,
        schedule_gate_bonus: float = 0.0,
    ) -> None:
        """Initialize the reward range and set the rewards and penalties.

        Args:
            illegal_action_penalty: Penalty for performing an illegal action. An action
                is illegal if ``action[0]`` is not in ``state["legal_actions"]``. This
                value should be negative (but is not required) and defaults to -5.
            update_cycle_penalty: Penalty given for incrementing a cycle. Since the
                ``Scheduling`` environment wats to create the shortest schedules,
                incrementing the cycle should be penalized. This value should be
                negative (but is not required) and defaults to -1.
            schedule_gate_bonus: Reward gained for successfully scheduling a gate. This
                value should be positive (but is not required) and defaults to 0.
        """
        self._illegal_action_penalty = check_real(
            illegal_action_penalty, "illegal_action_penalty"
        )
        self._update_cycle_penalty = check_real(
            update_cycle_penalty, "update_cycle_penalty"
        )
        self._schedule_gate_bonus = check_real(
            schedule_gate_bonus, "schedule_gate_bonus"
        )
        self._set_reward_range()

        warn_if_positive(self._illegal_action_penalty, "illegal_action_penalty")
        warn_if_positive(self._update_cycle_penalty, "update_cycle_penalty")
        warn_if_negative(self._schedule_gate_bonus, "schedule_gate_bonus")

    def compute_reward(
        self,
        *,
        old_state: SchedulingState,
        action: NDArray[np.int_],
        new_state: SchedulingState,
    ) -> float:
        """Compute a reward, based on the new state, and the given action. Specifically
        the 'legal_actions' actions array.

        Args:
            old_state: State of the ``Scheduling`` environment before the current
                action.
            action: Action that has just been taken.
            new_state: Updated state of the ``Scheduling`` environment.

        Returns:
            The reward for this action. If the action is illegal, then the reward is
            `illegal_action_penalty`. If the action is legal, and increments the cycle,
            then the reward is `update_cycle_penalty`. Otherwise, the reward is
            `schedule_gate_bonus`.
        """
        if action[1] != 0:
            return self._update_cycle_penalty

        if self._is_illegal(action, old_state):
            return self._illegal_action_penalty

        return self._schedule_gate_bonus

    @staticmethod
    def _is_illegal(action: NDArray[np.int_], old_state: SchedulingState) -> bool:
        """Check if the given action is illegal. An action is illegal if ``action[0]``
        is not in ``old_state["legal_actions"]``.

        Args:
            action: Action that has just been taken.
            old_state: State of the ``Scheduling`` before the current action.

        Returns:
            Boolean value stating whether this action was illegal.
        """
        gate_to_schedule = action[0]
        return not old_state.circuit_info.legal[gate_to_schedule]

    def _set_reward_range(self) -> None:
        """Set the reward range."""
        l_bound = -float("inf")
        if (
            self._illegal_action_penalty >= 0
            and self._update_cycle_penalty >= 0
            and self._schedule_gate_bonus >= 0
        ):
            l_bound = 0

        u_bound = float("inf")
        if (
            self._illegal_action_penalty <= 0
            and self._update_cycle_penalty <= 0
            and self._schedule_gate_bonus <= 0
        ):
            u_bound = 0

        self._reward_range = (l_bound, u_bound)


class EpisodeRewarder(Rewarder):
    """Rewarder for the ``Scheduling`` environment, which only gives a reward at
    the end of the episode or when an illegal action is taken.
    """

    def __init__(
        self, illegal_action_penalty: float = -5.0, update_cycle_penalty: float = -1.0
    ) -> None:
        """Initialize the reward range and set the rewards and penalties.

        Args:
            illegal_action_penalty: Penalty for performing an illegal action. An action
                is illegal if ``action[0]`` is not in ``state["legal_actions"]``. This
                value should be negative (but is not required) and defaults to -5.
            update_cycle_penalty: Penalty given for incrementing a cycle. Since the
                ``Scheduling`` environment wats to create the shortest schedules,
                incrementing the cycle should be penalized. This value should be
                negative (but is not required) and defaults to -1.
        """
        self._illegal_action_penalty = check_real(
            illegal_action_penalty, "illegal_action_penalty"
        )
        self._update_cycle_penalty = check_real(
            update_cycle_penalty, "update_cycle_penalty"
        )
        self._set_reward_range()

        warn_if_positive(self._illegal_action_penalty, "illegal_action_penalty")
        warn_if_positive(self._update_cycle_penalty, "update_cycle_penalty")

    def compute_reward(
        self,
        *,
        old_state: SchedulingState,
        action: NDArray[np.int_],
        new_state: SchedulingState,
    ) -> float:
        """Compute a reward, based on the new state, and the given action.

        Args:
            old_state: State of the ``Scheduling`` environment before the current
                action.
            action: Action that has just been taken.
            new_state: Updated state of the ``Scheduling`` environment.

        Returns:
            The reward for this action. If the action is illegal, then the reward is
            `illegal_action_penalty`. If the action is legal, but the episode is not yet
            done, then the reward is 0. Otherwise, the reward is
            `update_cycle_penalty`x`current cycle`.
        """
        if action[1] == 0 and self._is_illegal(action, old_state):
            return self._illegal_action_penalty

        if not new_state.is_done():
            return 0

        reward = 0
        for gate_idx, scheduled_cycle in enumerate(new_state.circuit_info.schedule):
            gate = new_state.circuit_info.encoded[gate_idx]
            finished = scheduled_cycle + new_state.gates[gate.name].cycle_length
            reward = min(reward, self._update_cycle_penalty * finished)

        return reward

    @staticmethod
    def _is_illegal(action: NDArray[np.int_], old_state: SchedulingState) -> bool:
        """Check if the given action is illegal. An action is illegal if ``action[0]``
        is not in ``old_state["legal_actions"]``.

        Args:
            action: Action that has just been taken.
            old_state: State of the ``Scheduling`` before the current action.

        Returns:
            Boolean value stating whether this action was illegal.
        """
        gate_to_schedule = action[0]
        return not old_state.circuit_info.legal[gate_to_schedule]

    def _set_reward_range(self) -> None:
        """Set the reward range."""
        l_bound = -float("inf")
        if self._illegal_action_penalty >= 0 and self._update_cycle_penalty >= 0:
            l_bound = 0

        u_bound = float("inf")
        if self._illegal_action_penalty <= 0 and self._update_cycle_penalty <= 0:
            u_bound = 0

        self._reward_range = (l_bound, u_bound)



"""Specific environments of this RL Gym in the Quantum domain. This package
contains the ``InitialMapping`` and ``Scheduling`` environments, which model their
respective OpenQL passes.
"""
from qgym.envs.initial_mapping.initial_mapping import InitialMapping
from qgym.envs.routing import Routing
from qgym.envs.scheduling.scheduling import Scheduling

__all__ = ["InitialMapping", "Routing", "Scheduling"]



"""Module containing the environment, rewarders, visualizer and other utils for the
scheduling problem of OpenQL.
"""
from qgym.envs.scheduling.machine_properties import MachineProperties
from qgym.envs.scheduling.rulebook import CommutationRulebook
from qgym.envs.scheduling.scheduling import Scheduling
from qgym.envs.scheduling.scheduling_rewarders import BasicRewarder, EpisodeRewarder
from qgym.envs.scheduling.scheduling_state import SchedulingState

__all__ = [
    "Scheduling",
    "SchedulingState",
    "MachineProperties",
    "CommutationRulebook",
    "BasicRewarder",
    "EpisodeRewarder",
]



"""This module contains some vanilla Rewarders for the ``InitialMapping`` environment.

Usage:
    The rewarders in this module can be customized by initializing the rewarders with
    different values.

    .. code-block:: python

        from qgym.envs.initial_mapping.initial_mapping_rewarders import BasicRewarder

        rewarder = BasicRewarder(
            illegal_action_penalty = -1,
            reward_per_edge = 5,
            penalty_per_edge: = -2,
            )

    After initialization, the rewarders can be given to the ``InitialMapping``
    environment.

.. note::
    When implementing custom rewarders, they should inherit from ``qgym.Rewarder``.
    Furthermore, they must implement the ``compute_reward`` method. Which takes as input
    the old state, the new state and the given action. See the documentation of the
    ``qgym.envs.initial_mapping.initial_mapping`` module for more information on
    the state and action space.

"""
import numpy as np
from numpy.typing import NDArray

from qgym.envs.initial_mapping.initial_mapping_state import InitialMappingState
from qgym.templates import Rewarder
from qgym.utils.input_validation import check_real, warn_if_negative, warn_if_positive


class BasicRewarder(Rewarder):
    """Basic rewarder for the ``InitialMapping`` environment."""

    __slots__ = (
        "_illegal_action_penalty",
        "_reward_per_edge",
        "_penalty_per_edge",
        "_reward_range",
    )

    def __init__(
        self,
        illegal_action_penalty: float = -100,
        reward_per_edge: float = 5,
        penalty_per_edge: float = -1,
    ) -> None:
        """Initialize the reward range and set the rewards and penalties.

        Args:
            illegal_action_penalty: Penalty for performing an illegal action. An action
                is illegal if the action contains a virtual or physical qubit that has
                already been mapped. This value should be negative (but is not required)
                and defaults to -100.
            reward_per_edge: Reward gained per 'good' edge in the interaction graph. An
                edge is 'good' if the mapped edge overlaps with an edge of the
                connection graph. This value should be positive (but is not required)
                and defaults to 5.
            penalty_per_edge: Penalty given per 'bad' edge in the interaction graph. An
                edge is 'bad' if the edge is mapped and is not 'good'. This value should
                be negative (but is not required) and defaults to -1.
        """
        self._illegal_action_penalty = check_real(
            illegal_action_penalty, "illegal_action_penalty"
        )
        self._reward_per_edge = check_real(reward_per_edge, "reward_per_edge")
        self._penalty_per_edge = check_real(penalty_per_edge, "penalty_per_edge")
        self._set_reward_range()

        warn_if_positive(self._illegal_action_penalty, "illegal_action_penalty")
        warn_if_negative(self._reward_per_edge, "reward_per_edge")
        warn_if_positive(self._penalty_per_edge, "penalty_per_edge")

    def compute_reward(
        self,
        *,
        old_state: InitialMappingState,
        action: NDArray[np.int_],
        new_state: InitialMappingState,
    ) -> float:
        """Compute a reward, based on the new state, and the given action.

        Specifically the connection graph, interaction graphs and mapping are used.

        Args:
            old_state: State of the ``InitialMapping`` before the current action.
            action: Action that has just been taken.
            new_state: Updated state of the ``InitialMapping``.

        Returns:
            The reward for this action. If the action is illegal, then the reward is
            `illegal_action_penalty`. If the action is legal, then the reward is the
            *total* number of 'good' edges times `reward_per_edge` plus the *total*
            number of 'bad' edges times `penalty_per_edge`.
        """
        if self._is_illegal(action, old_state):
            return self._illegal_action_penalty

        return self._compute_state_reward(new_state)

    def _compute_state_reward(self, state: InitialMappingState) -> float:
        """Compute the value of the mapping defined by the input state.

        Args:
            state: The state to compute the value of.

        Returns:
            The reward value of this state.
        """
        reward = 0.0
        for flat_idx in state.graphs["interaction"]["matrix"].nonzero()[0]:
            interaction_i, interaction_j = divmod(flat_idx, state.n_nodes)
            mapped_interaction_i = state.mapping_dict.get(interaction_i, None)
            mapped_interaction_j = state.mapping_dict.get(interaction_j, None)
            if mapped_interaction_i is None or mapped_interaction_j is None:
                continue
            if state.graphs["connection"]["matrix"][
                mapped_interaction_i, mapped_interaction_j
            ]:
                reward += self._reward_per_edge
            else:
                reward += self._penalty_per_edge

        return reward / 2  # divide by two due to double counting of edges

    @staticmethod
    def _is_illegal(action: NDArray[np.int_], old_state: InitialMappingState) -> bool:
        """Check if the given action is illegal.

        An action is considered illegal if qubits are mapped multiple times.

        Args:
            action: Action that has just been taken.
            old_state: State of the ``InitialMapping`` before the current action.

        Returns:
            Whether this action is valid for the given state.
        """
        return (
            action[0] in old_state.mapped_qubits["physical"]
            or action[1] in old_state.mapped_qubits["logical"]
        )

    def _set_reward_range(self) -> None:
        """Set the reward range."""
        l_bound = -float("inf")
        if (
            self._reward_per_edge >= 0
            and self._penalty_per_edge >= 0
            and self._illegal_action_penalty >= 0
        ):
            l_bound = 0

        u_bound = float("inf")
        if (
            self._reward_per_edge <= 0
            and self._penalty_per_edge <= 0
            and self._illegal_action_penalty <= 0
        ):
            u_bound = 0

        self._reward_range = (l_bound, u_bound)


class SingleStepRewarder(BasicRewarder):
    """Rewarder for the ``InitialMapping`` environment, which gives a reward based on
    the improvement in the current step.
    """

    def compute_reward(
        self,
        *,
        old_state: InitialMappingState,
        action: NDArray[np.int_],
        new_state: InitialMappingState,
    ) -> float:
        """Compute a reward, based on the new state, and the given action.

        Specifically the connection graph, interaction graphs and mapping are used.

        Args:
            old_state: State of the ``InitialMapping`` before the current action.
            action: Action that has just been taken.
            new_state: Updated state of the ``InitialMapping``.

        Returns:
            The reward for this action. If the action is illegal, then the reward is
            `illegal_action_penalty`. If the action is legal, then the reward is the
            number of 'good' edges times `reward_per_edge` plus the number of 'bad'
            edges times `penalty_per_edge` created by the *this* action.
        """
        if self._is_illegal(action, old_state):
            return self._illegal_action_penalty

        return self._compute_state_reward(new_state) - self._compute_state_reward(
            old_state
        )


class EpisodeRewarder(BasicRewarder):
    """Rewarder for the ``InitialMapping`` environment, which only gives a reward at
    the end of the episode or when an illegal action is taken.
    """

    def compute_reward(
        self,
        *,
        old_state: InitialMappingState,
        action: NDArray[np.int_],
        new_state: InitialMappingState,
    ) -> float:
        """Compute a reward, based on the new state, and the given action.

        Specifically the connection graph, interaction graphs and mapping are used.

        Args:
            old_state: State of the ``InitialMapping`` before the current action.
            action: Action that has just been taken.
            new_state: Updated state of the ``InitialMapping``.

        Returns:
            The reward for this action. If the action is illegal, then the reward is
            `illegal_action_penalty`. If the action is legal, but the mapping is not yet
            finished, then the reward is 0. If the action is legal, and the mapping is
            finished, then the reward is the number of 'good' edges times
            `reward_per_edge` plus the number of 'bad' edges times `penalty_per_edge`.
        """
        if self._is_illegal(action, old_state):
            return self._illegal_action_penalty

        if len(new_state.mapped_qubits["physical"]) != new_state.n_nodes:
            return 0

        return self._compute_state_reward(new_state)



"""This module contains a class used for rendering the ``Scheduling`` environment."""
from __future__ import annotations

from typing import Dict, cast

import numpy as np
import pygame
from numpy.typing import NDArray

from qgym.custom_types import Gate
from qgym.envs.scheduling.scheduling_state import SchedulingState
from qgym.templates.visualiser import RenderData, Visualiser
from qgym.utils.visualisation.colors import BLACK, BLUE, DARK_BLUE, WHITE
from qgym.utils.visualisation.typing import Font
from qgym.utils.visualisation.wrappers import write_text


class SchedulingVisualiser(Visualiser):
    """Visualiser class for the ``Scheduling`` environment."""

    def __init__(self, render_mode: str, initial_state: SchedulingState) -> None:
        """Init of the ``SchedulingVisualiser``.

        Args:
            initial_state: ``SchedulingState`` object containing the initial state of
                the environment to visualise.
            render_mode: If 'human' open a ``pygame`` screen visualizing the step. If
                'rgb_array', return an RGB array encoding of the rendered frame on each
                render call.
        """
        # Rendering data
        self.offset = {"x-axis": 100, "y-axis": 0}
        colors = {
            "gate_fill": BLUE,
            "gate_outline": DARK_BLUE,
            "gate_text": WHITE,
            "background": WHITE,
            "qubits": BLACK,
            "y-axis": BLACK,
        }
        self.render_data = RenderData(
            screen=self._start_screen(
                "Scheduling Environment", render_mode, (1500, 800)
            ),
            font=self._start_font(),
            colors=colors,
            render_mode=render_mode,
        )

        n_qubits = initial_state.machine_properties.n_qubits
        gate_height = (self.screen_height - self.offset["y-axis"]) / n_qubits

        longest_gate = 0
        for n_cycles in initial_state.machine_properties.gates.values():
            longest_gate = max(longest_gate, n_cycles)

        self.gate_size_info = {"height": gate_height, "longest": longest_gate}

        # define attributes that are set later
        self._cycle_width = 0.0

    def render(self, state: SchedulingState) -> None | NDArray[np.int_]:
        """Render the current state using pygame.

        Args:
            state: State to render.

        Raises:
            ValueError: If an unsupported mode is provided.

        Returns:
            Result of rendering, based on `render_mode`.
        """
        pygame.time.delay(50)

        self.screen.fill(self.colors["background"])
        self._draw_y_axis(state.machine_properties.n_qubits)

        self._cycle_width = (self.screen_width - self.offset["x-axis"]) / (
            state.cycle + self.gate_size_info["longest"]
        )

        for gate_idx, scheduled_cycle in enumerate(state.circuit_info.schedule):
            if scheduled_cycle != -1:
                gate = state.circuit_info.encoded[gate_idx]
                gate_cycle_length = cast(
                    Dict[int, int], state.machine_properties.gates
                )[gate.name]
                gate_name = state.utils.gate_encoder.decode_gates(gate.name)
                self._draw_scheduled_gate(
                    gate, scheduled_cycle, gate_cycle_length, gate_name
                )

        return self._display()

    def _draw_y_axis(self, n_qubits: int) -> None:
        """Draw the y-axis of the display.

        Args:
            n_qubits: Number of qubits of the machine.
        """
        screen = self.screen
        for i in range(n_qubits):
            pos = (
                self.offset["x-axis"] / 2,
                self.gate_size_info["height"] * (n_qubits - i - 0.5),
            )
            write_text(screen, self.font["axis"], f"Q{i}", pos, self.colors["y-axis"])

    def _draw_scheduled_gate(
        self, gate: Gate, scheduled_cycle: int, gate_cycle_length: int, gate_name: str
    ) -> None:
        """Draw a gate on the screen.

        Args:
            gate: Gate to draw.
            scheduled_cycle: Cycle the gate is scheduled.
            gate_cycle_length: Length of the gate in terms of machine cycles.
            gate_name: Name of the gate.
        """
        self._draw_gate_block(gate_name, gate_cycle_length, gate.q1, scheduled_cycle)
        if gate.q1 != gate.q2:
            self._draw_gate_block(
                gate_name, gate_cycle_length, gate.q2, scheduled_cycle
            )

    def _draw_gate_block(
        self, gate_name: str, gate_cycle_length: int, qubit: int, scheduled_cycle: int
    ) -> None:
        """Draw a single block of a gate (gates can consist of 1 or 2 blocks).

        Args:
            gate_name: Name of the gate.
            gate_cycle_length: Length of the gate in terms of machine cycles.
            qubit: Qubit in which the gate acts.
            scheduled_cycle: Cycle in which the gate is scheduled.
        """
        screen = self.screen
        gate_width = self._cycle_width * gate_cycle_length
        gate_box_size = (0.98 * gate_width, 0.98 * self.gate_size_info["height"])

        box_pos = (
            self.screen_width - scheduled_cycle * self._cycle_width - gate_width,
            self.screen_height
            - qubit * self.gate_size_info["height"]
            - self.offset["y-axis"]
            - self.gate_size_info["height"],
        )
        gate_box = pygame.Rect(box_pos, gate_box_size)

        pygame.draw.rect(screen, self.colors["gate_fill"], gate_box, border_radius=5)
        pygame.draw.rect(
            screen, self.colors["gate_outline"], gate_box, width=2, border_radius=5
        )
        write_text(
            screen=screen,
            font=self.font["gate"],
            text=gate_name.upper(),
            pos=gate_box.center,
            color=self.colors["gate_text"],
        )

    def _start_font(self) -> dict[str, Font]:
        """Start the ``pygame`` fonts for the gate and axis font.

        Returns:
            ``pygame`` fonts for the gate and axis font.
        """
        pygame.font.init()
        return {
            "gate": pygame.font.SysFont("Arial", 12),
            "axis": pygame.font.SysFont("Arial", 30),
        }



"""This module contains the ``RoutingState`` class.

This ``RoutingState``represents the ``State`` of the ``Routing`` environment.

Usage:
    >>> from qgym.envs.routing.routingstate import RoutingState
    >>> import networkx as nx
    >>> connection_graph = nx.grid_graph((3,3))
    >>> state = RoutingState(
    >>>             max_interaction_gates = 100,
    >>>             max_observation_reach = 20,
    >>>             connection_graph = connection_graph,
    >>>             observe_legal_surpasses = True,
    >>>             observe_connection_graph = False,
    >>>             )
"""
from __future__ import annotations

from collections import deque
from typing import Any, Dict

import networkx as nx
import numpy as np
from numpy.typing import NDArray

import qgym.spaces
from qgym.templates.state import State

# pylint: disable=too-many-instance-attributes


class RoutingState(State[Dict[str, NDArray[np.int_]], NDArray[np.int_]]):
    """The ``RoutingState`` class."""

    def __init__(
        self,
        *,
        max_interaction_gates: int,
        max_observation_reach: int,
        connection_graph: nx.Graph,
        observe_legal_surpasses: bool,
        observe_connection_graph: bool,
    ) -> None:
        """Init of the ``RoutingState`` class.

        Args:
            max_interaction_gates: Sets the maximum amount of gates in the
                interaction_circuit, when a new interaction_circuit is generated.
            max_observation_reach: Sets a cap on the maximum amount of gates the agent
                can see ahead when making an observation. When bigger than
                max_interaction_gates the agent will always see all gates ahead in an
                observation
            connection_graph: ``networkx`` graph representation of the QPU topology.
                Each node represents a physical qubit and each edge represents a
                connection in the QPU topology.
            observe_legal_surpasses: If ``True`` a boolean array of length
                max_observation_reach indicating whether the gates ahead can be
                executed, will be added to the `observation_space`.
            observe_connection_graph: If ``True``, the connection_graph will be
                incorporated in the `observation_space`. Reason to set it ``False`` is:
                QPU-topology doesn't change, hence an agent could infer the topology
                from the training data without needing to explicitly add it to the
                observations. This reduced the size `observation_space`.
        """
        self.steps_done = 0
        """Number of steps done since the last reset."""
        self.connection_graph = connection_graph
        """``networkx`` graph representation of the QPU topology. Each node represents a
        physical qubit and each edge represents a connection in the QPU topology.
        """
        self.max_interaction_gates = max_interaction_gates
        """Sets the maximum amount of gates in the interaction_circuit, when a new
        interaction_circuit is generated.
        """
        number_of_gates = self.rng.integers(1, self.max_interaction_gates + 1)
        self.interaction_circuit = self.generate_random_interaction_circuit(
            number_of_gates
        )
        """An array of 2-tuples of integers, where every tuple represents a, not
        specified, gate acting on the two qubits labeled by the integers in the tuples.
        """
        self.mapping = np.arange(self.n_qubits)
        """Array of which each index represents a logical qubit and each value
        represents a physical qubit.
        """
        self.position: int = 0
        """An integer representing before which gate in the interaction_circuit the
        agent currently is.
        """
        self.max_observation_reach = max_observation_reach
        """An integer that sets a cap on the maximum amount of gates the agent can see
        ahead when making an observation. When bigger than max_interaction_gates the
        agent will always see all gates ahead in an observation.
        """
        self.observe_legal_surpasses = observe_legal_surpasses

        if observe_connection_graph:
            self.connection_matrix = nx.to_numpy_array(
                connection_graph, dtype=np.int_
            ).flatten()

        # Keep track of at what position which swap_gate is inserted
        self.swap_gates_inserted: deque[tuple[int, int, int]] = deque()
        """A deque of 3-tuples of integers, to register which gates to insert and where.
        Every tuple (g, q1, q2) represents the insertion of a SWAP-gate acting on
        logical qubits q1 and q2 before gate g in the interaction_circuit.
        """

    def reset(
        self,
        *,
        seed: int | None = None,
        interaction_circuit: NDArray[np.int_] | None = None,
        **_kwargs: Any,
    ) -> RoutingState:
        """Reset the state (in place) and load a new (random) initial state.

        To be used after an episode is finished.

        Args:
            seed: Seed for the random number generator, should only be provided
                (optionally) on the first reset call, i.e., before any learning is done.
            circuit: Optional list of tuples of ints that the interaction gates via the
                qubits the gates are acting on.
            _kwargs: Additional options to configure the reset.

        Returns:
            Self.
        """
        if seed is not None:
            self.seed(seed)

        if interaction_circuit is None:
            number_of_gates = self.rng.integers(1, self.max_interaction_gates + 1)
            self.interaction_circuit = self.generate_random_interaction_circuit(
                number_of_gates
            )
        else:
            interaction_circuit = np.array(interaction_circuit)
            max_gates = self.max_interaction_gates
            if (
                interaction_circuit.ndim != 2
                or interaction_circuit.shape[0] > max_gates
                or interaction_circuit.shape[1] != 2
            ):
                msg = "'interaction_circuit' should have be an ArrayLike with shape "
                msg += (
                    "(n_interactions,2), where n_interactions<=max_interaction_gates."
                )
                raise ValueError(msg)
            self.interaction_circuit = interaction_circuit

        # Reset position, counters
        self.position = 0
        self.steps_done = 0

        # resetting swap_gates_inserted and mapping
        self.swap_gates_inserted = deque()
        self.mapping = np.arange(self.n_qubits, dtype=np.int_)

        return self

    def obtain_info(
        self,
    ) -> dict[str, int | deque[tuple[int, int, int]] | NDArray[np.int_]]:
        """Obtain additional information of the current state.

        Returns:
            Dictionary containing optional debugging info for the current state.
        """
        return {
            "Steps done": self.steps_done,
            "Position": self.position,
            "Interaction gates ahead": np.array(
                [
                    self.interaction_circuit[idx]
                    for idx in range(self.position, len(self.interaction_circuit))
                ]
            ),
            "Number of swaps inserted": len(self.swap_gates_inserted),
            "Swap gates inserted": self.swap_gates_inserted,
        }

    def update_state(self, action: NDArray[np.int_]) -> RoutingState:
        """Update the state (in place) of this environment using the given action.

        Args:
            action: If action[0]==0 a SWAP-gate applied to qubits action[1], action[2]
                will be registered in the swap_gates_inserted-deque at the current
                position, if action[0]==1 the first observed gate will be surpassed.

        Returns:
            Self.
        """
        # Increase the step number
        self.steps_done += 1

        surpass, qubit1, qubit2 = action
        # surpass current_gate if legal
        if surpass and self.is_legal_surpass(*self.interaction_circuit[self.position]):
            self.position += 1

        # elif insert swap-gate if legal
        elif not surpass and self.is_legal_swap(qubit1, qubit2):
            self._place_swap_gate(qubit1, qubit2)
            self._update_mapping(qubit1, qubit2)

        return self

    def create_observation_space(self) -> qgym.spaces.Dict:
        """Create the corresponding observation space.

        Returns:
            Observation space in the form of a ``qgym.spaces.Dict`` space containing:

            * ``qgym.spaces.MultiDiscrete`` space representing the interaction gates
              ahead of current position.
            * ``qgym.spaces.MultiDiscrete`` space representing the current mapping of
              logical onto physical qubits
        """
        interaction_gates_ahead = qgym.spaces.MultiDiscrete(
            np.full(2 * self.max_observation_reach, self.n_qubits + 1)
        )
        mapping = qgym.spaces.MultiDiscrete(np.full(self.n_qubits, self.n_qubits))

        observation_kwargs: dict[str, Any]
        observation_kwargs = {
            "interaction_gates_ahead": interaction_gates_ahead,
            "mapping": mapping,
        }

        if hasattr(self, "connection_matrix"):
            observation_kwargs["connection_graph"] = qgym.spaces.Box(
                low=0,
                high=np.iinfo(np.int64).max,
                shape=(self.n_qubits * self.n_qubits,),
                dtype=np.int64,
            )

        if self.observe_legal_surpasses:
            observation_kwargs["is_legal_surpass"] = qgym.spaces.MultiBinary(
                self.max_observation_reach
            )

        return qgym.spaces.Dict(observation_kwargs)

    def obtain_observation(
        self,
    ) -> dict[str, NDArray[np.int_]]:
        """Observe the current state.

        Returns:
            Observation based on the current state.
        """
        # construct interaction_gates_ahead
        gate_slice = slice(self.position, self.position + self.max_observation_reach)
        interaction_gates_ahead = self.interaction_circuit[gate_slice]
        n_pad = self.max_observation_reach - len(interaction_gates_ahead)
        interaction_gates_ahead = np.pad(
            interaction_gates_ahead,
            ((0, n_pad), (0, 0)),
            constant_values=self.n_qubits,
        )

        observation = {
            "interaction_gates_ahead": interaction_gates_ahead.flatten(),
            "mapping": self.mapping,
        }

        if hasattr(self, "connection_matrix"):
            observation["connection_graph"] = self.connection_matrix

        if self.observe_legal_surpasses:
            is_legal_surpass = np.array(
                [self.is_legal_surpass(*gate) for gate in interaction_gates_ahead]
            )
            observation["is_legal_surpass"] = is_legal_surpass

        return observation

    def is_done(self) -> bool:
        """Checks if the current state is in a final state.

        Returs: Boolean value stating whether we are in a final state.
        """
        return self.position == len(self.interaction_circuit)

    def _place_swap_gate(
        self,
        logical_qubit1: int,
        logical_qubit2: int,
    ) -> None:
        """Place a swap gate at the current position with the given logical qubits."""
        self.swap_gates_inserted.append((self.position, logical_qubit1, logical_qubit2))

    def is_legal_swap(
        self,
        logical_qubit1: int,
        logical_qubit2: int,
    ) -> bool:
        """Checks whether a swap of two logical qubits is legal.

        Returns:
            Boolean value state whether the swap is allowed in the connection graph..
        """
        physical_qubit1 = self.mapping[logical_qubit1]
        physical_qubit2 = self.mapping[logical_qubit2]
        return (logical_qubit1 != logical_qubit2) and (
            (physical_qubit1, physical_qubit2) in self.connection_graph.edges
        )

    def is_legal_surpass(
        self,
        logical_qubit1: int,
        logical_qubit2: int,
    ) -> bool:
        """Checks whether a surpass of the current gate ahead is legal.

        Args:
            logical_qubit1: First qubit of the interaction.
            logical_qubit2: Second qubit of the logical interaction.

        Returns:
            A boolean value stating whether a connection gate with the two qubits can be
            executed with the current mapping and connection graph.
        """
        try:
            physical_connection = self.mapping[[logical_qubit1, logical_qubit2]]
        except IndexError:
            # The only logical qubits that are out of index, are those of padded gates.
            return True
        return physical_connection in self.connection_graph.edges

    def _update_mapping(
        self,
        logical_qubit1: int,
        logical_qubit2: int,
    ) -> None:
        """Updates mapping for a swap of two qubits."""
        physical_qubits = self.mapping[[logical_qubit1, logical_qubit2]]
        self.mapping[[logical_qubit2, logical_qubit1]] = physical_qubits

    def generate_random_interaction_circuit(self, n_gates: int) -> NDArray[np.int_]:
        """Generate a random interaction circuit.

        Returns:
            A randomly generated interaction circuit.
        """
        circuit = np.zeros((n_gates, 2), dtype=int)
        for idx in range(n_gates):
            circuit[idx] = self.rng.choice(
                np.arange(self.n_qubits), size=2, replace=False
            )

        return circuit

    @property
    def n_qubits(self) -> int:
        """Number of qubits in the `connection_graph`."""
        return int(self.connection_graph.number_of_nodes())



"""This module contains the ``RandomCircuitGenerator``, which can be used to generate
random circuits. Here, a quantum circuit is a ``List`` with ``Gate`` objects.
"""
from __future__ import annotations

import numpy as np
from numpy.random import Generator, default_rng

from qgym.custom_types import Gate


class RandomCircuitGenerator:
    """Generates random circuits in the form of a ``List`` with ``Gate`` objects.

    Example circuit:
        >>> circuit = [Gate("prep", 0,0), Gate("prep", 1,1), Gate("cnot", 0,1)]
    """

    def __init__(
        self, n_qubits: int, max_gates: int, rng: Generator | None = None
    ) -> None:
        """Initialize the ``RandomCircuitGenerator``.

        Args:
            n_qubits: Number of qubits of the circuit.
            max_gates: Maximum number of gates that a circuit may have.
            rng: Optional random number generator.
        """
        self.n_qubits = n_qubits
        self.max_gates = max_gates
        self._rng = rng

    @property
    def rng(self) -> Generator:
        """Return the random number generator of this ``RandomCircuitGenerator``.

        If none is set yet, this will generate a new one, using
        ``numpy.random.default_rng``.
        """
        if self._rng is None:
            self._rng = default_rng()
        return self._rng

    @rng.setter
    def rng(self, rng: Generator) -> None:
        self._rng = rng

    def generate_circuit(
        self, n_gates: str | int = "random", mode: str = "default"
    ) -> list[Gate]:
        """Generate a random quantum circuit.

        Args:
            n_gates: If "random", then a circuit of random length will be made. If an
                ``int`` is given, a circuit of length ``min(n_gates, max_gates)`` will
                be made.
            mode: If mode is "default", a circuit will be generated containing the
                'prep', 'x', 'y', 'z', 'cnot' and 'measure' gates. If mode is
                "workshop", a simpler circuit containing just 'h', 'cnot' and `measure`
                gates will be generated.

        Returns:
            A randomly generated quantum circuit.
        """
        n_gates = self._parse_n_gates(n_gates)

        circuit: list[Gate] = [Gate("", -1, -1)] * n_gates

        if mode.lower() == "default":
            gate_names = ["x", "y", "z", "cnot", "measure"]
            probabilities = [0.16, 0.16, 0.16, 0.5, 0.02]
        elif mode.lower() == "workshop":
            gate_names = ["x", "y", "cnot", "measure"]
            probabilities = [0.2, 0.2, 0.5, 0.1]
        else:
            raise ValueError("Unknown mode, choose 'default' or 'workshop'.")

        for idx in range(n_gates):
            name = self.rng.choice(gate_names, p=probabilities)

            if name == "cnot":
                qubit1, qubit2 = self.rng.choice(
                    np.arange(self.n_qubits), size=2, replace=False
                )
            else:
                qubit1 = self.rng.integers(self.n_qubits)
                qubit2 = qubit1

            circuit[idx] = Gate(name, qubit1, qubit2)

        # If mode is default, the circuit should start by initializing the qubits
        if mode.lower() == "default":
            for qubit in range(self.n_qubits):
                circuit[qubit] = Gate("prep", qubit, qubit)

        return circuit

    def _parse_n_gates(self, n_gates: int | str) -> int:
        """Parse `n_gates`.

        Args:
            n_gates: If n_gates is "random", generate a number between 1 and
                `max_gates`. If n_gates is an ``int``, return the minimum of `n_gates`
                and `max_gates`.
        """
        if isinstance(n_gates, str):
            if n_gates.lower().strip() == "random":
                return self.rng.integers(self.n_qubits, self.max_gates, endpoint=True)

            raise ValueError(f"Unknown flag {n_gates}, choose from 'random'.")

        if isinstance(n_gates, int):
            return min(n_gates, self.max_gates)

        msg = f"n_gates should be of type int or str, but was of type {type(n_gates)}."
        raise ValueError(msg)

    def __repr__(self) -> str:
        """Create a string representation of the RandomCircuitGenerator."""
        text = f"{self.__class__.__name__}({self.n_qubits}, {self.max_gates}, "
        text += f"{self.rng})"
        return text



"""This module contains a class used for rendering a ``Routing`` environment."""
from __future__ import annotations

from typing import Any, Sequence

import networkx as nx
import numpy as np
import pygame
from networkx import Graph
from numpy.typing import NDArray

from qgym.envs.routing.routing_state import RoutingState
from qgym.templates.visualiser import RenderData, Visualiser
from qgym.utils.visualisation.colors import BLACK, BLUE, GRAY, RED, WHITE
from qgym.utils.visualisation.typing import Font, Surface
from qgym.utils.visualisation.wrappers import (
    draw_point,
    draw_wide_line,
    shade_rect,
    write_text,
)


class RoutingVisualiser(Visualiser):
    """Visualiser class for the ``Routing`` environment."""

    def __init__(self, render_mode: str, connection_graph: Graph) -> None:
        """Init of the ``RoutingVisualiser``.

        Args:
            connection_graph: ``networkx.Graph`` representation of the connection graph.
            render_mode: If 'human' open a ``pygame`` screen visualizing the step. If
                'rgb_array', return an RGB array encoding of the rendered frame on each
                render call.
        """
        # Rendering data
        colors = {
            "node": BLUE,
            "node_labels": WHITE,
            "edge": BLACK,
            "good gate": BLUE,
            "bad gate": RED,
            "passed gate": GRAY,
            "text": BLACK,
            "circuit lines": BLACK,
            "background": WHITE,
            "hidden": BLACK,
            "mapping": WHITE,
        }
        self.render_data = RenderData(
            screen=self._start_screen("Routing Environment", render_mode, (1600, 700)),
            font=self._setup_fonts(),
            colors=colors,
            render_mode=render_mode,
        )

        self.subscreens = self._start_subscreens(self.screen)

        # Save everything we need to know about the connection graph
        self.graph = {
            "render_positions": self._get_render_positions(connection_graph),
            "nodes": connection_graph.nodes,
            "edges": connection_graph.edges,
        }

    def _start_subscreens(
        self, screen: Surface, padding: int = 20
    ) -> tuple[Surface, Surface]:
        """Initialize the subscreens.

        Args:
            screen: The parent screen.
            padding: The padding to be used in between the subscreens.
        """
        position_circuit = (0, self.header_spacing)
        width_circuit = self.screen_width * 0.75 - 0.5 * padding
        height_circuit = self.screen_height - self.header_spacing
        rect_circuit = pygame.Rect(position_circuit, (width_circuit, height_circuit))
        subscreen_circuit = screen.subsurface(rect_circuit)

        position_graph = (width_circuit + padding, self.header_spacing)
        width_graph = self.screen_width * 0.25 - 0.5 * padding
        height_graph = self.screen_height - self.header_spacing
        rect_graph = pygame.Rect(position_graph, (width_graph, height_graph))
        subscreen_graph = screen.subsurface(rect_graph)
        return subscreen_circuit, subscreen_graph

    def render(self, state: RoutingState) -> None | NDArray[np.int_]:
        """Render the current state using ``pygame``.

        Args:
            state: State to render.

        Raises:
            ValueError: When an invalid mode is provided.

        Returns:
            If `render_mode` is 'human' returns show the current step at using a
            ``pygame`` screen. If `render_mode` is 'rgb_array' returns a RGB array
            encoding of the rendered image.
        """
        self.screen.fill(self.colors["background"])

        self._draw_connection_graph(self.subscreens[1])
        self._draw_interaction_circuit(state, self.subscreens[0])

        self._draw_header("Interaction Circuit", self.subscreens[0])
        self._draw_header("Connection Graph", self.subscreens[1])

        return self._display()

    def _draw_interaction_circuit(self, state: RoutingState, screen: Surface) -> None:
        """Draw the interaction circuit on the interaction circuit subscreen.

        Args:
            state: Current state.
            screen: (Sub)screen to draw the circuit on.
        """
        x_text = screen.get_width() * 0.05
        x_left = screen.get_width() * 0.1
        x_right = screen.get_width() * 0.95
        y_distance = screen.get_height() / (state.n_qubits)
        y_lines = y_distance * (0.5 + np.arange(state.n_qubits))
        dx_gates = (x_right - x_left) / state.max_interaction_gates
        x_gates = x_left + dx_gates * (0.5 + np.arange(state.max_interaction_gates))

        self._draw_circuit_lines(
            screen, x_text=x_text, x_left=x_left, x_right=x_right, y_lines=y_lines
        )
        self._draw_interaction_gates(
            screen, state=state, x_gates=x_gates, y_lines=y_lines
        )
        self._draw_observation_reach(
            screen, state=state, x_left=x_left, x_right=x_right
        )
        self._draw_mapping(screen, state=state, x_gates=x_gates, y_lines=y_lines)

    def _draw_circuit_lines(
        self,
        screen: Surface,
        *,
        x_text: float,
        x_left: float,
        x_right: float,
        y_lines: NDArray[np.float_],
    ) -> None:
        """Draw the circuit lines on the 'circuit' screen and label them.

        Args:
            screen: (Sub)screen to draw the circuit lines on.
            x_text: x coordinate of the labels of the circuit lines.
            x_left: Left most x coordinate of the circuit lines.
            x_right: Right most x coordinate of the circuit lines.
            y_lines: Array of y coordinates of the circuit lines.
        """
        for qubit_idx, y_line in enumerate(y_lines):
            draw_wide_line(
                screen,
                self.colors["circuit lines"],
                (x_left, y_line),
                (x_right, y_line),
            )
            write_text(
                screen,
                font=self.font["header"],
                text=f"Q{qubit_idx}",
                pos=(x_text, y_line),
                color=self.colors["text"],
            )

    def _draw_interaction_gates(
        self,
        screen: Surface,
        *,
        state: RoutingState,
        x_gates: NDArray[np.float_],
        y_lines: NDArray[np.float_],
    ) -> None:
        """Draw the interaction gates on the 'circuit' screen.

        Args:
            screen: (Sub)screen to draw the interactions on.
            state: ``RoutingState`` to draw the interaction gates.
            x_gates: Array of x coordinates of the swap gates.
            y_lines: Array of y coordinates of the circuit lines.
        """
        for i, (qubit1, qubit2) in enumerate(state.interaction_circuit):
            physical_qubit1, physical_qubit2 = state.mapping[[qubit1, qubit2]]
            if i < state.position:
                color = self.colors["passed gate"]
            elif (physical_qubit1, physical_qubit2) in self.graph["edges"]:
                color = self.colors["good gate"]
            else:
                color = self.colors["bad gate"]

            point1 = (x_gates[i], y_lines[qubit1])
            point2 = (x_gates[i], y_lines[qubit2])
            draw_wide_line(screen, color, point1, point2)
            draw_point(screen, point1, color)
            draw_point(screen, point2, color)

    def _draw_observation_reach(
        self, screen: Surface, *, state: RoutingState, x_left: float, x_right: float
    ) -> None:
        """Draw shades on the 'circuit' screen to show the observation size.

        Args:
            screen: (Sub)screen to draw the observation reach on.
            state: Current state to draw the observation reach of.
            x_left: Left most x coordinate of the circuit lines.
            x_right: Right most x coordinate of the circuit lines.
        """
        dx_gates = (x_right - x_left) / state.max_interaction_gates

        shade_left_width = (state.position + 0.5) * dx_gates
        shade_height = screen.get_height()
        shade_rect(
            screen,
            size=(shade_left_width, shade_height),
            pos=(x_left - 0.25 * dx_gates, 0),
            color=self.colors["hidden"],
            alpha=128,
        )

        shade_right_width = (
            state.max_interaction_gates - state.max_observation_reach - state.position
        ) * dx_gates
        if shade_right_width > 0:
            shade_rect(
                screen,
                size=(shade_right_width, shade_height),
                pos=(x_right - shade_right_width, 0),
                color=self.colors["hidden"],
                alpha=128,
            )

    def _draw_mapping(
        self,
        screen: Surface,
        *,
        state: RoutingState,
        x_gates: NDArray[np.float_],
        y_lines: NDArray[np.float_],
    ) -> None:
        """Draw the mapping on the 'circuit' screen.

        Args:
            screen: (Sub)screen to draw the mapping on.
            state: ``RoutingState`` to draw the mapping of.
            x_gates: Array of x coordinates of the swap gates.
            y_lines: Array of y coordinates of the circuit lines.
        """
        dx_gates = x_gates[1] - x_gates[0]
        mapping = np.arange(state.n_qubits, dtype=int)
        starting_idx = 0
        for i in range(state.position):
            old_mapping = mapping.copy()
            starting_idx, mapping, n_swaps = self._update_mapping(
                mapping=mapping,
                swap_gates_inserted=state.swap_gates_inserted,
                position=i,
                starting_idx=starting_idx,
            )

            # Draw the mapping
            x_mapping = x_gates[i] - 0.5 * dx_gates
            for physical_qubit, y_logical_qubit, is_changed in zip(
                mapping, y_lines, old_mapping != mapping
            ):
                write_text(
                    screen,
                    font=self.font["mapping_emph" if is_changed else "mapping"],
                    text=str(physical_qubit),
                    pos=(x_mapping, y_logical_qubit),
                    color=self.colors["mapping"],
                )

            write_text(
                screen,
                font=self.font["n_swaps"],
                text=str(n_swaps),
                pos=(x_mapping, 1.3 * y_lines[0] - 0.3 * y_lines[1]),
                color=self.colors["mapping"],
            )

    def _update_mapping(
        self,
        *,
        mapping: NDArray[np.int_],
        swap_gates_inserted: Sequence[tuple[int, int, int]],
        position: int,
        starting_idx: int = 0,
    ) -> tuple[int, NDArray[np.int_], int]:
        """Update the mapping to conform to the current position.

        Args:
            mapping: Mapping of the previous position
            swap_gates_inserted: List of swap gates inserted.
            position: Position in the interaction circuit where the mapping must be
                made.
            starting_idx: Index of the last swap gate of the previous position, i.e.,
                position in the `swap_gates_inserted` sequence.

        Returns:
            Tuple with a starting index for the next position, the updated mapping and
            number of swap gates used since the last mapping.
        """
        n_swaps = 0
        if starting_idx >= len(swap_gates_inserted):
            return starting_idx, mapping, n_swaps

        for i in range(starting_idx, len(swap_gates_inserted)):
            swap_position, qubit1, qubit2 = swap_gates_inserted[i]
            if position != swap_position:
                break

            n_swaps += 1
            mapping[[qubit1, qubit2]] = mapping[[qubit2, qubit1]]
        return i, mapping, n_swaps

    def _draw_connection_graph(self, screen: Surface) -> None:
        """Draw the connection graph on the graph subscreen.

        Args:
            screen: (Sub)screen to draw the connection graph on.
        """
        for node_u, node_v in self.graph["edges"]:
            pos_u = self.graph["render_positions"][node_u]
            pos_v = self.graph["render_positions"][node_v]
            draw_wide_line(screen, self.colors["edge"], pos_u, pos_v)

        for label, pos in self.graph["render_positions"].items():
            draw_point(screen, pos, self.colors["node"], 20)
            write_text(
                screen, self.font["graph"], str(label), pos, self.colors["node_labels"]
            )

    def _draw_header(self, text: str, screen: Surface) -> None:
        """Draw a header above a subscreen.

        Args:
            text: Text of the header.
            screen: Subscreen to draw the header of.
        """
        pygame_text = self.font["header"].render(text, True, self.colors["text"])
        offset = screen.get_offset()
        rect = screen.get_rect(topleft=offset)
        text_center = (rect.center[0], rect.y - self.header_spacing / 2)
        text_position = pygame_text.get_rect(center=text_center)
        self.screen.blit(pygame_text, text_position)

    def _get_render_positions(
        self, graph: nx.Graph, padding: int = 20
    ) -> dict[Any, NDArray[np.float_]]:
        """Give the positions of the nodes of a graph on a given screen.

        Args:
            graph: Graph of which the node positions must be determined.
            screen: the subscreen on which the graph will be drawn.

        Returns:
            Dictionary where the keys are the names of the nodes, and the values are the
            coordinates of these nodes.
        """
        node_positions: dict[Any, NDArray[np.float_]]
        node_positions = nx.spring_layout(graph, threshold=1e-6)

        # Scale and move the node positions to be centered on the graph subscreen
        width_graph_screen = self.screen_width * 0.25 - 0.5 * padding
        height_graph_screen = self.screen_height - self.header_spacing
        size = np.array([width_graph_screen, height_graph_screen])
        for node, position in node_positions.items():
            node_positions[node] = position * 0.45 * size + 0.5 * size

        return node_positions

    def _setup_fonts(self) -> dict[str, Font]:
        """Setup the fonts for rendering with pygame."""
        pygame.font.init()
        return {
            "header": pygame.font.SysFont("Arial", 30),
            "circuit": pygame.font.SysFont("Arial", 24),
            "mapping": pygame.font.SysFont("Arial", 22),
            "mapping_emph": pygame.font.SysFont("Arial", 24, bold=True, italic=True),
            "n_swaps": pygame.font.SysFont("Arial", 28),
            "graph": pygame.font.SysFont("Arial", 24),
        }

    @property
    def header_spacing(self) -> float:
        """Header spacing."""
        return 30 / 3 * 4



"""This module contains the ``SchedulingState`` class.

This ``SchedulingState``represents the ``State`` of the ``Scheduling`` environment.
"""
from __future__ import annotations

from typing import Any, Dict, Set, Union, cast

import numpy as np
from numpy.typing import NDArray

import qgym.spaces
from qgym.custom_types import Gate
from qgym.envs.scheduling.machine_properties import MachineProperties
from qgym.envs.scheduling.rulebook import CommutationRulebook
from qgym.envs.scheduling.scheduling_dataclasses import (
    CircuitInfo,
    GateInfo,
    SchedulingUtils,
)
from qgym.templates.state import State
from qgym.utils.random_circuit_generator import RandomCircuitGenerator


class SchedulingState(
    State[Dict[str, Union[NDArray[np.int_], NDArray[np.int8]]], NDArray[np.int_]]
):
    """The ``SchedulingState`` class."""

    def __init__(
        self,
        *,
        machine_properties: MachineProperties,
        max_gates: int,
        dependency_depth: int,
        random_circuit_mode: str,
        rulebook: CommutationRulebook,
    ) -> None:
        """Init of the ``SchedulingState`` class.

        Args:
            machine_properties: A ``MachineProperties`` object.
            max_gates: Maximum number of gates allowed in a circuit.
            dependency_depth: Number of dependencies given in the observation.
                Determines the shape of the `dependencies` observation, which has the
                shape (dependency_depth, max_gates).
            random_circuit_mode: Mode for the random circuit generator. The mode can be
                'default' or 'workshop'.
            rulebook: ``CommutationRulebook`` describing the commutation rules.
        """
        self.steps_done = 0
        """Number of steps done since the last reset."""
        self.cycle = 0
        """Current 'machine' cycle."""
        self.machine_properties = machine_properties
        """``MachineProperties`` class containing machine properties and limitations."""
        self.utils = SchedulingUtils(
            random_circuit_generator=RandomCircuitGenerator(
                machine_properties.n_qubits, max_gates, rng=self.rng
            ),
            random_circuit_mode=random_circuit_mode,
            rulebook=rulebook,
            gate_encoder=machine_properties.encode(),
        )
        """``SchedulingUtils`` dataclass with a random circuit generator, commutation
        rulebook and a gate encoder.
        """

        # At the start no gates should be excluded
        gate_cycle_lengths = cast(Dict[int, int], machine_properties.gates)
        not_in_same_cycle = cast(
            Dict[int, Set[int]], machine_properties.not_in_same_cycle
        )
        self.gates = {
            gate_name: GateInfo(
                cycle_length=gate_cycle_lengths[gate_name],
                not_in_same_cycle=not_in_same_cycle[gate_name],
            )
            for gate_name in gate_cycle_lengths
        }
        """Dictionary with gate names as keys and ``GateInfo`` dataclasses as values."""
        self.busy = np.zeros(machine_properties.n_qubits, dtype=int)
        """Amount of cycles that a qubit is still busy (zero if available). Used
        internally for the hardware limitations.
        """

        # Generate a circuit
        circuit = self.utils.random_circuit_generator.generate_circuit(
            mode=self.utils.random_circuit_mode
        )

        self.circuit_info = CircuitInfo(
            encoded=self.utils.gate_encoder.encode_gates(circuit),
            names=np.empty(max_gates, dtype=int),
            acts_on=np.empty((2, max_gates), dtype=int),
            legal=np.empty(max_gates, dtype=np.int8),
            dependencies=np.empty((dependency_depth, max_gates), dtype=int),
            schedule=np.full(len(circuit), -1, dtype=int),
            blocking_matrix=self.utils.rulebook.make_blocking_matrix(circuit),
        )
        """``CircuitInfo`` dataclass containing the encoded circuit and attributes used
        to update the state.
        """

        self._update_dependencies()
        self._update_episode_constant_observations()
        self._update_legal_actions()

    def _update_dependencies(self) -> None:
        """Compute and update the dependencies array of the current state."""
        self.circuit_info.dependencies = np.zeros_like(self.circuit_info.dependencies)

        for gate_idx, blocking_row in enumerate(self.circuit_info.blocking_matrix):
            blocking_gates = blocking_row[gate_idx:].nonzero()[0]
            for depth in range(
                min(self.circuit_info.dependencies.shape[0], blocking_gates.shape[0])
            ):
                self.circuit_info.dependencies[depth, gate_idx] = blocking_gates[depth]

    def _update_episode_constant_observations(self) -> None:
        """Update episode constant observations `gate_names` and `acts_on`.

        Based on the circuit of the current episode.
        """
        self.circuit_info.names = np.zeros_like(self.circuit_info.names)
        self.circuit_info.acts_on = np.zeros_like(self.circuit_info.acts_on)

        for gate_idx, gate in enumerate(self.circuit_info.encoded):
            self.circuit_info.names[gate_idx] = gate.name
            self.circuit_info.acts_on[0, gate_idx] = gate.q1
            self.circuit_info.acts_on[1, gate_idx] = gate.q2

    def _update_legal_actions(self) -> None:
        """Check which actions are legal based on the scheduled qubits.

        An action is legal if the gate could be scheduled based on the machine
        properties and commutation rules.
        """
        self.circuit_info.legal = np.zeros_like(self.circuit_info.legal)
        for gate_idx, (gate_name, qubit1, qubit2) in enumerate(
            self.circuit_info.encoded
        ):
            # Set all gates, which have not been scheduled to True
            if self.circuit_info.schedule[gate_idx] == -1:
                self.circuit_info.legal[gate_idx] = True

            # Check if there is a non-scheduled dependent gate

            dependent_gates = self.circuit_info.dependencies[:, gate_idx]
            if np.count_nonzero(dependent_gates) > 0:
                self.circuit_info.legal[gate_idx] = False
                continue

            # Check if the qubits are busy
            if self.busy[qubit1] > 0 or self.busy[qubit2] > 0:
                self.circuit_info.legal[gate_idx] = False
                continue

            # Check if gates should be excluded
            if self.gates[gate_name].exclude > 0:
                self.circuit_info.legal[gate_idx] = False
                continue

    def create_observation_space(self) -> qgym.spaces.Dict:
        """Create the corresponding observation space.

        Returns:
            Observation space in the form of a ``qgym.spaces.Dict`` space containing:

            * ``qgym.spaces.MultiBinary`` space representing the legal actions. If
              the value at index $i$ determines if gate number $i$ can be scheduled
              or not.
            * ``qgym.spaces.MultiDiscrete`` space representing the integer encoded
              gate names.
            * ``qgym.spaces.MultiDiscrete`` space representing the interaction of
              each gate (q1 and q2).
            * ``qgym.spaces.MultiDiscrete`` space representing the first $n$ gates
              that must be scheduled before this gate.
        """
        max_gates = len(self.circuit_info.legal)
        n_gates = self.machine_properties.n_gates
        n_qubits = self.machine_properties.n_qubits
        dependency_depth = self.circuit_info.dependencies.shape[0]

        legal_actions_space = qgym.spaces.MultiBinary(max_gates, rng=self.rng)
        gate_names_space = qgym.spaces.MultiDiscrete(
            np.full(max_gates, n_gates + 1), rng=self.rng
        )
        acts_on_space = qgym.spaces.MultiDiscrete(
            np.full(2 * max_gates, n_qubits + 1), rng=self.rng
        )
        dependencies_space = qgym.spaces.MultiDiscrete(
            np.full(dependency_depth * max_gates, max_gates), rng=self.rng
        )

        observation_space = qgym.spaces.Dict(
            rng=self.rng,
            legal_actions=legal_actions_space,
            gate_names=gate_names_space,
            acts_on=acts_on_space,
            dependencies=dependencies_space,
        )
        return observation_space

    def obtain_observation(
        self,
    ) -> dict[str, NDArray[np.int_] | NDArray[np.int8]]:
        """Obtain an observation based on the current state.

        Returns:
            Observation based on the current state.
        """
        return {
            "gate_names": self.circuit_info.names,
            "acts_on": self.circuit_info.acts_on.flatten(),
            "dependencies": self.circuit_info.dependencies.flatten(),
            "legal_actions": self.circuit_info.legal,
        }

    def is_done(self) -> bool:
        """Determine if the state is done or not.

        Returns:
            Boolean value stating whether we are in a final state.
        """
        return all(self.circuit_info.schedule != -1)

    def obtain_info(self) -> dict[str, int | NDArray[np.int_]]:
        """Obtain additional information.

        Returns:
            Optional debugging info for the current state.
        """
        return {
            "Steps done": self.steps_done,
            "Cycle": self.cycle,
            "Schedule": self.circuit_info.schedule,
        }

    def update_state(self, action: NDArray[np.int_]) -> SchedulingState:
        """Update the state of this environment using the given action.

        Args:
            action: First entry determines a gate to schedule, the second entry
                increases the cycle if nonzero.

        Returns:
            Self.
        """
        # Increase the step number
        self.steps_done += 1

        # Increase the cycle if the action is given
        if action[1]:
            self._increment_cycle()
            return self

        # Schedule the gate if it is allowed
        gate_to_schedule = action[0]
        if self.circuit_info.legal[gate_to_schedule]:
            self._schedule_gate(gate_to_schedule)
        return self

    def reset(
        self,
        *,
        seed: int | None = None,
        circuit: list[Gate] | None = None,
        **_kwargs: Any,
    ) -> SchedulingState:
        """Reset the state and load a new (random) initial state.

        To be used after an episode is finished.

        Args:
            seed: Seed for the random number generator, should only be provided
                (optionally) on the first reset call, i.e., before any learning is done.
            circuit: Optional list of a circuit for the next episode, each entry in the
                list should be a ``Gate``. When a circuit is give, no random circuit
                will be generated.
            _kwargs: Additional options to configure the reset.

        Returns:
            Self.
        """
        if seed is not None:
            self.seed(seed)

        # Reset counters
        self.steps_done = 0
        self.cycle = 0

        # Amount of cycles that a qubit is still busy (zero if available)
        self.busy = np.zeros_like(self.busy)

        # At the start no gates should be excluded
        for gate_info in self.gates.values():
            gate_info.reset()

        # Generate a circuit if None is given
        self.circuit_info.reset(circuit, self.utils)

        self._update_dependencies()
        self._update_episode_constant_observations()
        self._update_legal_actions()

        return self

    def _increment_cycle(self) -> None:
        """Increment the cycle and update the state accordingly."""
        self.cycle += 1

        # Reduce the amount of cycles each qubit is busy
        self.busy[self.busy > 0] -= 1

        # Exclude gates that should start at the same time
        for gate_name, gate_info in self.gates.items():
            if gate_info.exclude_next_cycle:
                gate_info.exclude_next_cycle = False
                self._exclude_gate(gate_name)

        # Decrease the amount of cycles to exclude a gate and skip gates where the
        # cycle becomes 0 (as it no longer should be excluded)
        for gate_info in self.gates.values():
            if gate_info.exclude > 0:
                gate_info.exclude -= 1

        self._update_legal_actions()

    def _exclude_gate(self, gate_name: int) -> None:
        """Exclude a gate from the 'legal_actions' for 'gate_cycle_length' cycles.

        Args:
            gate_name: integer encoding of the name of the gate.
        """
        gate_cycle_length = self.gates[gate_name].cycle_length
        self.gates[gate_name].exclude = gate_cycle_length

    def _schedule_gate(self, gate_idx: int) -> None:
        """Schedule a gate in the current cycle and update the state accordingly.

        Args:
            gate_idx: Index of the gate to schedule.
        """
        gate = self.circuit_info.encoded[gate_idx]

        # add the gate to the schedule
        self.circuit_info.schedule[gate_idx] = self.cycle

        self.busy[gate.q1] = self.gates[gate.name].cycle_length
        self.busy[gate.q2] = self.gates[gate.name].cycle_length

        for gate_to_exclude in self.gates[gate.name].not_in_same_cycle:
            self._exclude_gate(gate_to_exclude)

        if gate.name in self.machine_properties.same_start:
            self.gates[gate.name].exclude_next_cycle = True

        # Update "dependencies" observation
        self.circuit_info.blocking_matrix[:gate_idx, gate_idx] = False
        self._update_dependencies()
        self._update_legal_actions()



"""The qgym package consist of gyms and tools used for reinforcement learning (RL)
environments in the Quantum domain. It's main purpose is to easily create RL
environments for the different passes of the OpenQL framework, by simply initializing an
environment class. This abstraction of the environment allows RL developers to develop
RL agents to improve the OpenQL framework, without requiring prior knowledge of OpenQL.


Example:
    We want to create an environment for the OpenQL pass of initial mapping for a system
    with a QPU topology of 3x3. Using the ``qgym`` package this becomes:

    .. code-block:: python

        from qgym.envs import InitialMapping
        env = InitialMapping(0.5, connection_grid_size=(3,3))

    We can then use the environment in the code block above to train a stable baseline
    RL agent using the following code:

    .. code-block:: python

        from stable_baselines3 import PPO
        model = PPO("MultiInputPolicy", env, verbose=1)
        model.learn(int(1e5))

"""

__version__ = "0.0.1"



"""This module contains the ``InitialMappingState`` class.

This ``InitialMappingState`` represents the ``State`` of the ``InitialMapping``
environment.

Usage:
    >>> from qgym.envs.initial_mapping.initial_mapping_state import InitialMappingState
    >>> import networkx as nx
    >>> connection_graph = nx.grid_graph((3,3))
    >>> state = InitialMappingState(connection_graph, 0.5)

"""
from __future__ import annotations

from copy import deepcopy
from typing import Any, Dict, cast

import networkx as nx
import numpy as np
from numpy.typing import NDArray

from qgym import spaces
from qgym.templates.state import State


class InitialMappingState(State[Dict[str, NDArray[np.int_]], NDArray[np.int_]]):
    """The ``InitialMappingState`` class."""

    __slots__ = (
        "steps_done",
        "graphs",
        "mapping",
        "mapping_dict",
        "mapped_qubits",
    )

    def __init__(
        self, connection_graph: nx.Graph, interaction_graph_edge_probability: float
    ) -> None:
        """Init of the ``InitialMappingState`` class.

        Args:
            connection_graph: ``networkx`` graph representation of the QPU topology.
                Each node represents a physical qubit and each edge represents a
                connection in the QPU topology.
            interaction_graph_edge_probability: Probability that an edge between any
                pair of qubits in the random interaction graph exists. The interaction
                graph will have the same amount of nodes as the connection graph. Nodes
                without any interactions can be seen as 'null' nodes. Must be a value in
                the range $[0,1]$.
        """
        # Create a random connection graph with `n_nodes` and with edges existing with
        # probability `interaction_graph_edge_probability` (nodes without connections
        # can be seen as 'null' nodes)
        interaction_graph = nx.fast_gnp_random_graph(
            connection_graph.number_of_nodes(),
            interaction_graph_edge_probability,
        )

        self.steps_done: int = 0
        """Number of steps done since the last reset."""

        self.graphs = {
            "connection": {
                "graph": deepcopy(connection_graph),
                "matrix": nx.to_numpy_array(connection_graph, dtype=np.int8),
            },
            "interaction": {
                "graph": deepcopy(interaction_graph),
                "matrix": nx.to_numpy_array(interaction_graph, dtype=np.int8).flatten(),
                "edge_probability": interaction_graph_edge_probability,
            },
        }
        """Dictionary containing the graph and matrix representations of the both the
        interaction graph and connection graph.
        """
        self.mapping = np.full(self.n_nodes, self.n_nodes, dtype=np.int_)
        """Array of which the index represents a physical qubit, and the value a virtual
        qubit. A value of ``n_nodes + 1`` represents the case when nothing is mapped to
        the physical qubit yet.
        """
        self.mapping_dict: dict[int, int] = {}
        """Dictionary that maps logical qubits (keys) to physical qubits (values)."""
        self.mapped_qubits: dict[str, set[int]] = {"physical": set(), "logical": set()}
        """Dictionary with two sets containing mapped physical and logical qubits."""

    def create_observation_space(self) -> spaces.Dict:
        """Create the corresponding observation space.

        Returns:
            Observation space in the form of a ``qgym.spaces.Dict`` space containing:

            * ``qgym.spaces.MultiDiscrete`` space representing the mapping.
            * ``qgym.spaces.MultiBinary`` representing the interaction matrix.
        """
        mapping_space = spaces.MultiDiscrete(
            nvec=[self.n_nodes + 1] * self.n_nodes, rng=self.rng
        )
        interaction_matrix_space = spaces.MultiBinary(self.n_nodes**2, rng=self.rng)
        return spaces.Dict(
            rng=self.rng,
            mapping=mapping_space,
            interaction_matrix=interaction_matrix_space,
        )

    def reset(
        self,
        *,
        seed: int | None = None,
        interaction_graph: nx.Graph | None = None,
        **_kwargs: Any,
    ) -> InitialMappingState:
        """Reset the state and set a new interaction graph.

        To be used after an episode is finished.

        Args:
            seed: Seed for the random number generator, should only be provided
                (optionally) on the first reset call i.e., before any learning is done.
            interaction_graph: Interaction graph to be used for the next iteration, if
            ``None`` a random interaction graph will be created.
            _kwargs: Additional options to configure the reset.

        Returns:
            (self) New initial state.
        """
        if seed is not None:
            self.seed(seed)

        # Reset the state
        if interaction_graph is None:
            self.graphs["interaction"]["graph"] = nx.fast_gnp_random_graph(
                self.n_nodes, self.graphs["interaction"]["edge_probability"]
            )
        else:
            self.graphs["interaction"]["graph"] = deepcopy(interaction_graph)

        self.graphs["interaction"]["matrix"] = nx.to_numpy_array(
            self.graphs["interaction"]["graph"], dtype=np.int8
        ).flatten()

        self.steps_done = 0
        self.mapping = np.full(self.n_nodes, self.n_nodes)
        self.mapping_dict = {}
        self.mapped_qubits = {"physical": set(), "logical": set()}

        return self

    def add_random_edge_weights(self) -> None:
        """Add random weights to the connection graph and interaction graph."""
        for node1, node2 in self.graphs["connection"]["graph"].edges():
            weight = self.rng.gamma(2, 2) / 4
            self.graphs["connection"]["graph"].edges[node1, node2]["weight"] = weight
        self.graphs["connection"]["matrix"] = nx.to_numpy_array(
            self.graphs["connection"]["graph"]
        )

        for node1, node2 in self.graphs["interaction"]["graph"].edges():
            weight = self.rng.gamma(2, 2) / 4
            self.graphs["interaction"]["graph"].edges[node1, node2]["weight"] = weight
        self.graphs["interaction"]["matrix"] = nx.to_numpy_array(
            self.graphs["interaction"]["graph"]
        ).flatten()

    def update_state(self, action: NDArray[np.int_]) -> InitialMappingState:
        """Update the state (in place) of this environment using the given action.

        Args:
            action: Mapping action to be executed.

        Returns:
            Self.
        """
        # Increase the step number
        self.steps_done += 1

        # update state based on the given action
        physical_qubit, logical_qubit = action

        if (
            physical_qubit in self.mapped_qubits["physical"]
            or logical_qubit in self.mapped_qubits["logical"]
        ):
            return self

        self.mapping[physical_qubit] = logical_qubit
        self.mapping_dict[logical_qubit] = physical_qubit
        self.mapped_qubits["physical"].add(physical_qubit)
        self.mapped_qubits["logical"].add(logical_qubit)
        return self

    def obtain_observation(self) -> dict[str, NDArray[np.int_]]:
        """Obtain an observation based on the current state.

        Returns:
            Observation based on the current state.
        """
        return {
            "mapping": self.mapping,
            "interaction_matrix": self.graphs["interaction"]["matrix"],
        }

    def is_done(self) -> bool:
        """Determine if the state is done or not.

        Returns:
            Boolean value stating whether we are in a final state.
        """
        return bool(len(self.mapping_dict) == self.n_nodes)

    def obtain_info(self) -> dict[str, Any]:
        """Obtain additional information.

        Returns:
            Optional debugging info for the current state.
        """
        return {"Steps done": self.steps_done}

    @property
    def n_nodes(self) -> int:
        """The number of physical qubits."""
        return cast(int, self.graphs["connection"]["graph"].number_of_nodes())



r"""This module contains an environment for training an RL agent on the initial mapping
problem of OpenQL. The initial mapping problem is aimed at mapping virtual qubits of a
circuit to physical qubits that have a certain connection topology. The quantum circuit
is represented as an **interaction graph**, where each node represent a qubit and each
edge represent an interaction between two qubits as defined by the circuit (See the
example below). The QPU structure is called the **connection graph**. In the connection
graph each node represents a physical qubit and each edge represent a connection between
two qubits in the QPU.


.. code-block:: console

              QUANTUM CIRCUIT                        INTERACTION GRAPH
           ┌───┐               ┌───┐
    |q3>───┤ R ├───┬───────────┤ M ╞══                 q1 ────── q2
           └───┘   │           └───┘                            ╱
           ┌───┐ ┌─┴─┐         ┌───┐                           ╱
    |q2>───┤ R ├─┤ X ├───┬─────┤ M ╞══                        ╱
           └───┘ └───┘   │     └───┘                         ╱
           ┌───┐       ┌─┴─┐   ┌───┐                        ╱
    |q1>───┤ R ├───┬───┤ X ├───┤ M ╞══                     ╱
           └───┘   │   └───┘   └───┘                      ╱
           ┌───┐ ┌─┴─┐         ┌───┐                     ╱
    |q0>───┤ R ├─┤ X ├─────────┤ M ╞══                q3 ─────── q4
           └───┘ └───┘         └───┘



The goal is to create a mapping between the nodes of the interaction and connection
graph, such that for every edge in the interaction graph, there is an edge in the
connection graph. If this is impossible, then the number of mismatches should be
penalized.


State Space:
    The state space is described by a ``InitialMappingState`` with the following
    attributes:

    * `steps_done`: Number of steps done since the last reset.
    * `n_nodes`: Number of *physical* qubits.
    * `graphs`: Dictionary containing the graph and matrix representations of the both
      the interaction graph and connection graph.
    * `mapping`: Array of which the index represents a physical qubit, and the value a
      virtual qubit. A value of ``n_nodes + 1`` represents the case when nothing is
      mapped to the physical qubit yet.
    * `mapping_dict`: Dictionary that maps logical qubits (keys) to physical qubit
      (values).
    * `mapped_qubits`: Dictionary with a two Sets containing all mapped physical and
      logical qubits.

Observation Space:
    The observation space is a ``qgym.spaces.Dict`` with 2 entries:

    * `mapping`: The current state of the mapping.
    * `interaction_matrix`: The flattened adjacency matrix of the interaction graph.

Action Space:
    A valid action is a tuple of integers  $(i,j)$, such that  $0 \le i, j < n$, where
    $n$ is the number of physical qubits. The action  $(i,j)$ maps virtual qubit $j$ to
    phyiscal qubit $i$ when this action is legal. An action is legal when:

    #. virtual qubit $i$ has not been mapped to another physical qubit; and
    #. no other virual qubit has been mapped to physical qubit $j$.

Example 1:
    Creating an environment with a gridlike connection graph is done by executing the
    following code:

    >>> from qgym.envs.initial_mapping import InitialMapping
    >>> env = InitialMapping(0.5, connection_grid_size=(3,3))

    By default,  ``InitialMapping`` uses the ``BasicRewarder``. As an example, we would
    like to change the rewarder to the ``EpisodeRewarder``. This can be done in the
    following way:

    >>> from qgym.envs.initial_mapping import EpisodeRewarder
    >>> env.rewarder = EpisodeRewarder()


Example 2:
    In this example we use a custom connection graph depicted in the code block below.

    .. code-block:: console

        q1──────q0──────q2
                 │
                 │
                 │
                q3


    The graph has a non-gridlike structure.
    Such connection graphs can be given to the environment by giving an adjacency matrix
    representation of the graph, or a ``networkx`` representation of the graph. We will
    show the latter option in this example.

    .. code-block:: python

        import networkx as nx
        from qgym.envs.initial_mapping import InitialMapping

        # Create a networkx representation of the connection graph
        connection_graph = nx.Graph()
        connection_graph.add_edge(0, 1)
        connection_graph.add_edge(0, 2)
        connection_graph.add_edge(0, 3)

        # Initialize the environment with the custom connection graph
        env = InitialMapping(0.5, connection_graph=connection_graph)


"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterable, Mapping, cast

import networkx as nx
import numpy as np
from numpy.typing import ArrayLike, NDArray

import qgym.spaces
from qgym.envs.initial_mapping.initial_mapping_rewarders import BasicRewarder
from qgym.envs.initial_mapping.initial_mapping_state import InitialMappingState
from qgym.envs.initial_mapping.initial_mapping_visualiser import (
    InitialMappingVisualiser,
)
from qgym.templates import Environment, Rewarder
from qgym.utils.input_parsing import (
    parse_connection_graph,
    parse_rewarder,
    parse_visualiser,
)
from qgym.utils.input_validation import check_real

if TYPE_CHECKING:
    Gridspecs = list[int | Iterable[int]] | tuple[int | Iterable[int]]


class InitialMapping(Environment[Dict[str, NDArray[np.int_]], NDArray[np.int_]]):
    """RL environment for the initial mapping problem of OpenQL."""

    __slots__ = (
        "_rewarder",
        "_state",
        "observation_space",
        "action_space",
        "metadata",
        "_visualiser",
    )

    def __init__(
        self,
        interaction_graph_edge_probability: float,
        *,
        connection_graph: nx.Graph | None = None,
        connection_graph_matrix: ArrayLike | None = None,
        connection_grid_size: Gridspecs | None = None,
        rewarder: Rewarder | None = None,
        render_mode: str | None = None,
    ) -> None:
        """Initialize the action space, observation space, and initial states.
        Furthermore, the connection graph and edge probability for the random
        interaction graph of each episode is defined.

        The supported render modes of this environment are "human" and "rgb_array".

        Args:
            interaction_graph_edge_probability: Probability that an edge between any
                pair of qubits in the random interaction graph exists. The interaction
                graph will have the same amount of nodes as the connection graph. Nodes
                without any interactions can be seen as 'null' nodes. Must be a value in
                the range [0,1].
            connection_graph: ``networkx`` graph representation of the QPU topology.
                Each node represents a physical qubit and each node represents a
                connection in the QPU topology.
            connection_graph_matrix: Adjacency matrix representation of the QPU
                topology.
            connection_grid_size: Size of the connection graph when the connection graph
                has a grid topology. For more information on the allowed values and
                types, see ``networkx`` `grid_graph`_ documentation.
            rewarder: Rewarder to use for the environment. Must inherit from
                ``qgym.Rewarder``. If ``None`` (default), then ``BasicRewarder`` is
                used.
            render_mode: If 'human' open a ``pygame`` screen visualizing the step. If
                'rgb_array', return an RGB array encoding of the rendered frame on each
                render call.

        .. _grid_graph: https://networkx.org/documentation/stable/reference/generated/
            networkx.generators.lattice.grid_graph.html#grid-graph
        """
        # Check user input and parse it to a uniform format
        interaction_graph_edge_probability = check_real(
            interaction_graph_edge_probability,
            "interaction_graph_edge_probability",
            l_bound=0,
            u_bound=1,
        )
        connection_graph = parse_connection_graph(
            connection_graph, connection_graph_matrix, connection_grid_size
        )

        self._rewarder = parse_rewarder(rewarder, BasicRewarder)

        # Define internal attributes
        self._state = InitialMappingState(
            connection_graph, interaction_graph_edge_probability
        )
        self.observation_space = self._state.create_observation_space()
        # Define attributes defined in parent class
        self.action_space = qgym.spaces.MultiDiscrete(
            nvec=[self._state.n_nodes, self._state.n_nodes], rng=self.rng
        )

        self.metadata = {"render.modes": ["human", "rgb_array"]}
        self._visualiser = parse_visualiser(
            render_mode, InitialMappingVisualiser, [connection_graph]
        )

    def reset(
        self,
        *,
        seed: int | None = None,
        options: Mapping[str, Any] | None = None,
    ) -> tuple[dict[str, NDArray[np.int_]], dict[str, Any]]:
        """Reset the state and set a new interaction graph.

        To be used after an episode is finished.

        Args:
            seed: Seed for the random number generator, should only be provided
                (optionally) on the first reset call i.e., before any learning is done.
            return_info: Whether to receive debugging info. Default is ``False``.
            options: Mapping with keyword arguments with additional options for the
                reset. Keywords can be found in the description of
                ``InitialMappingState.reset()``

        Returns:
            Initial observation and debugging info.
        """
        # call super method for dealing with the general stuff
        return super().reset(seed=seed, options=options)

    def add_random_edge_weights(self) -> None:
        """Add random weights to the connection graph and interaction graph."""
        cast(InitialMappingState, self._state).add_random_edge_weights()



"""This module contains the ``MultiBinary`` space, i.e., an array of binary values.

Usage:
    >>> from qgym.spaces import MultiBinary
    >>> MultiBinary(10)
    MultiBinary(10)

"""
from __future__ import annotations

import gymnasium.spaces
import numpy as np
from numpy.random import Generator
from numpy.typing import ArrayLike


class MultiBinary(gymnasium.spaces.MultiBinary):
    """Multi-binary action/observation space for use in RL environments."""

    def __init__(
        self,
        n: ArrayLike | int,
        *,
        rng: Generator | None = None,
    ) -> None:
        """Initialize a multi-discrete space, i.e., multiple discrete intervals of given
        sizes.

        Args:
            n: ArrayLike containing integers representing the number of elements in the
                space.
            rng: Random number generator to be used in this space. If ``None``, a new
                random number generator will be constructed.
        """
        if isinstance(n, int):
            super().__init__(n)
        else:
            super().__init__(np.asarray(n))
        self._np_random = rng  # this overrides the default behaviour of the gym space



"""This module contains the ``Box`` space, i.e., a possibly open-ended interval in $n$
dimensions.

Usage:
    >>> from qgym.spaces import Box
    >>> Box(low=0, high=20, shape=(2, 3), dtype=int)
    Box([[0 0 0]
     [0 0 0]], [[20 20 20]
     [20 20 20]], (2, 3), int32)

"""
from __future__ import annotations

from numbers import Integral
from typing import Any, Sequence, SupportsFloat

import gymnasium.spaces
import numpy as np
from numpy.random import Generator
from numpy.typing import ArrayLike


class Box(gymnasium.spaces.Box):
    """An $n$-dimensional box space, i.e., collection of (possibly) open-ended
    intervals.
    """

    def __init__(
        self,
        low: SupportsFloat | ArrayLike,
        high: SupportsFloat | ArrayLike,
        shape: Sequence[int] | None = None,
        dtype: type[np.floating[Any]] | type[np.integer[Any]] = np.float_,
        *,
        rng: Generator | None = None,
    ) -> None:
        """Initialize a ``Box`` space, i.e., a possibly open-ended interval in $n$
        dimensions.

        Args:
            low: Either one lower bound for all intervals, or an ``NDArray`` with the
                shape given in `shape` with unique lower bounds for each interval.
            high: Either one upper bound for all intervals, or an ``NDArray`` with the
                shape given in `shape` with unique upper bounds for each interval.
            shape: ``Tuple`` containing the shape of the ``Box`` space.
            dtype: Type of the values in each interval.
            rng: Random number generator to be used in this space, if ``None`` a new one
                will be constructed.
        """
        if not isinstance(low, Integral):
            low = low.__float__() if hasattr(low, "__float__") else np.asarray(low)
        if not isinstance(high, Integral):
            high = high.__float__() if hasattr(high, "__float__") else np.asarray(high)
        super().__init__(low, high, shape=shape, dtype=dtype)
        self._np_random = rng  # this overrides the default behaviour of the gym space



"""Module containing the environment, rewarders, visualizer and other utils for the
routing problem of OpenQL.
"""
from qgym.envs.routing.routing import Routing
from qgym.envs.routing.routing_rewarders import (
    BasicRewarder,
    EpisodeRewarder,
    SwapQualityRewarder,
)
from qgym.envs.routing.routing_state import RoutingState

__all__ = [
    "Routing",
    "RoutingState",
    "BasicRewarder",
    "EpisodeRewarder",
    "SwapQualityRewarder",
]



"""This module will contain some vanilla Rewarders for the ``Routing`` environment.

Usage:
    The rewarders in this module can be customized by initializing the rewarders with
    different values.

    .. code-block:: python

        from qgym.envs.routing import BasicRewarder

        rewarder = BasicRewarder(
            illegal_action_penalty = -1,
            update_cycle_penalty = -2,
            schedule_gate_bonus: = 3,
            )

    After initialization, the rewarders can be given to the ``Routing`` environment.

.. note::
    When implementing custom rewarders, they should inherit from ``qgym.Rewarder``.
    Furthermore, they must implement the ``compute_reward`` method. Which takes as input
    the old state, the new state and the given action. See the documentation of the
    ``qgym.envs.routing.routing`` module for more information on the state and
    action space.

"""
import warnings

import numpy as np
from numpy.typing import NDArray

from qgym.envs.routing.routing_state import RoutingState
from qgym.templates import Rewarder
from qgym.utils.input_validation import check_real, warn_if_negative, warn_if_positive


class BasicRewarder(Rewarder):
    """RL Rewarder, for computing rewards on the ``RoutingState``."""

    def __init__(
        self,
        illegal_action_penalty: float = -50,
        penalty_per_swap: float = -10,
        reward_per_surpass: float = 10,
    ) -> None:
        """Set the rewards and penalties.

        Args:
            illegal_action_penalty: Penalty for performing an illegal action. An action
                is illegal if ``action[0]`` is not in ``state["legal_actions"]``. This
                value should be negative (but is not required) and defaults to -50.
            penalty_per_swap: Penalty for placing a swap. In general, we want to have as
                little swaps as possible. Therefore, this value should be negative and
                defaults to -10.
            reward_per_surpass: Reward given for surpassing a gate. In general, we want
                to have go to the end of the circuit as fast as possible. Therefore,
                this value should be positive and defaults to 10.
        """
        self._illegal_action_penalty = check_real(
            illegal_action_penalty, "illegal_action_penalty"
        )
        self._penalty_per_swap = check_real(penalty_per_swap, "penalty_per_swap")
        self._reward_per_surpass = check_real(reward_per_surpass, "reward_per_surpass")
        self._set_reward_range()

        warn_if_positive(self._illegal_action_penalty, "illegal_action_penalty")
        warn_if_positive(self._penalty_per_swap, "penalty_per_swap")
        warn_if_negative(self._reward_per_surpass, "reward_per_surpass")

    def compute_reward(
        self,
        *,
        old_state: RoutingState,
        action: NDArray[np.int_],
        new_state: RoutingState,
    ) -> float:
        """Compute a reward, based on the old state, new state, and the given action.

        Args:
            old_state: ``RoutingState`` before the current action.
            action: Action that has just been taken.
            new_state: ``RoutingState`` after the current action.

        Returns:
            The reward for this action.
        """

        if self._is_illegal(action, old_state):
            return self._illegal_action_penalty

        reward = old_state.position * self._reward_per_surpass
        reward += len(old_state.swap_gates_inserted) * self._penalty_per_swap
        reward += self._reward_per_surpass if action[0] else self._penalty_per_swap

        return reward

    def _is_illegal(self, action: NDArray[np.int_], old_state: RoutingState) -> bool:
        """Checks whether an action chosen by the agent is illegal.

        Returns:
            Boolean value stating whether the action was illegal or not.
        """
        if action[0]:
            qubit1, qubit2 = old_state.interaction_circuit[old_state.position]
            return not old_state.is_legal_surpass(qubit1, qubit2)

        return not old_state.is_legal_swap(action[1], action[2])

    def _set_reward_range(self) -> None:
        """Set the reward range."""
        l_bound = -float("inf")
        if (
            self._illegal_action_penalty >= 0
            and self._penalty_per_swap >= 0
            and self._reward_per_surpass >= 0
        ):
            l_bound = 0

        u_bound = float("inf")
        if (
            self._illegal_action_penalty <= 0
            and self._penalty_per_swap <= 0
            and self._reward_per_surpass <= 0
        ):
            u_bound = 0

        self._reward_range = (l_bound, u_bound)


class SwapQualityRewarder(BasicRewarder):
    """Rewarder for the ``Routing`` environment which takes swap qualities into account.

    The ``SwapQualityRewarder`` has an adjusted reward w.r.t. the ``BasicRewarder`` in
    the sense that good SWAPs give lower penalties and bad SWAPs give higher penalties.
    """

    def __init__(
        self,
        illegal_action_penalty: float = -50,
        penalty_per_swap: float = -10,
        reward_per_surpass: float = 10,
        good_swap_reward: float = 5,
    ) -> None:
        """Set the rewards and penalties and a flag.

        Args:
            illegal_action_penalty: Penalty for performing an illegal action. An action
                is illegal if ``action[0]`` is not in ``state["legal_actions"]``. This
                value should be negative (but is not required) and defaults to -50.
            penalty_per_swap: Penalty for placing a swap. In general, we want to have as
                little swaps as possible. Therefore, this value should be negative and
                defaults to -10.
            reward_per_surpass: Reward given for surpassing a gate. In general, we want
                to have go to the end of the circuit as fast as possible. Therefore,
                this value should be positive and defaults to 10.
            good_swap_reward: Reward given for placing a good swap. In general, we want
                to place as little swaps as possible. However, when they are good, the
                penalty for the placement should be surpressed. That happens with this
                reward. So, the value should be positive and smaller than the
                penalty_per_swap, in order not to get positive rewards for swaps,
                defaults to 5.
        """
        super().__init__(
            illegal_action_penalty=illegal_action_penalty,
            penalty_per_swap=penalty_per_swap,
            reward_per_surpass=reward_per_surpass,
        )

        self._good_swap_reward = check_real(good_swap_reward, "reward_per_good_swap")

        if not 0 <= self._good_swap_reward < -self._penalty_per_swap:
            warnings.warn("Good swaps should not result in positive rewards.")

        warn_if_negative(self._good_swap_reward, "reward_per_good_swap")

    def compute_reward(
        self,
        *,
        old_state: RoutingState,
        action: NDArray[np.int_],
        new_state: RoutingState,
    ) -> float:
        """Compute a reward, based on the old state, the given action and the new state.

        Specifically, the change in observation reach is used.

        Args:
            old_state: ``RoutingState`` before the current action.
            action: Action that has just been taken.
            new_state: ``RoutingState`` after the current action.

        Returns:
            The reward for this action. If the action is illegal, then the reward is
            `illegal_action_penalty`. If the action is legal, then the reward for a
            surpass is just reward_per_surpass. But, for a legal swap the reward
            adjusted with respect to the BasicRewarder. Namely, the penalty of a swap is
            reduced if it increases the observation_reach and the penalty is increased
            if the observation_reach is decreases.
        """
        if self._is_illegal(action, old_state):
            return self._illegal_action_penalty

        if action[0]:
            return self._reward_per_surpass

        return (
            self._penalty_per_swap
            + self._good_swap_reward
            * self._observation_enhancement_factor(old_state, new_state)
        )

    def _observation_enhancement_factor(
        self,
        old_state: RoutingState,
        new_state: RoutingState,
    ) -> float:
        """Calculates the change of the observation reach as an effect of a swap.

        Args:
            old_state: ``RoutingState`` before the current action.
            new_state: ``RoutingState`` after the current action.

        Returns:
            A fraction that expresses the procentual improvement w.r.t the `old_state`'s
            observation.
        """
        try:
            is_legal_surpass = old_state.obtain_observation()["is_legal_surpass"]
            old_executable_gates_ahead = int(is_legal_surpass.sum())

            is_legal_surpass = new_state.obtain_observation()["is_legal_surpass"]
            new_executable_gates_ahead = int(is_legal_surpass.sum())
        except KeyError as error:
            if not old_state.observe_legal_surpasses:
                msg = "observe_legal_surpasses needs to be True to compute"
                msg += "observation_enhancement_factor"
                raise ValueError(msg) from error
            raise error

        return (
            new_executable_gates_ahead - old_executable_gates_ahead
        ) / old_state.max_observation_reach

    def _set_reward_range(self) -> None:
        """Set the reward range."""
        l_bound = -float("inf")
        if (
            self._illegal_action_penalty >= 0
            and self._penalty_per_swap >= 0
            and self._reward_per_surpass >= 0
            and self._good_swap_reward >= 0
        ):
            l_bound = 0

        u_bound = float("inf")
        if (
            self._illegal_action_penalty <= 0
            and self._penalty_per_swap <= 0
            and self._reward_per_surpass <= 0
            and self._good_swap_reward <= 0
        ):
            u_bound = 0

        self._reward_range = (l_bound, u_bound)


class EpisodeRewarder(BasicRewarder):
    """Rewarder for the ``Routing`` environment, which only gives a reward after at
    the end of a full episode. The reward is the lowest for the lowest amount of SWAPs.
    This could be improved for setting for taking into account the fidelity of edges and
    scoring good and looking at what edges the circuit is executed.
    """

    def compute_reward(
        self,
        *,
        old_state: RoutingState,
        action: NDArray[np.int_],
        new_state: RoutingState,
    ) -> float:
        """Compute a reward, based on the new state, and the given action.

        Args:
            old_state: ``RoutingState`` before the current action.
            action: Action that has just been taken.
            new_state: ``RoutingState`` after the current action.

        Returns:
            If an action is illegal returns the `illegal_action_penalty`. If the episode
            is finished returns the reward calculated over the episode, otherwise
            returns 0.
        """
        if self._is_illegal(action, old_state):
            return self._illegal_action_penalty

        if not new_state.is_done():
            return 0

        return len(new_state.swap_gates_inserted) * self._penalty_per_swap

    def _set_reward_range(self) -> None:
        """Set the reward range."""
        l_bound = -float("inf")
        if self._illegal_action_penalty >= 0 and self._penalty_per_swap >= 0:
            l_bound = 0

        u_bound = float("inf")
        if self._illegal_action_penalty <= 0 and self._penalty_per_swap <= 0:
            u_bound = 0

        self._reward_range = (l_bound, u_bound)



"""This module contains function which parse user input.

With parsing we mean that the user input is validated and transformed to a predictable
format. In this way, user can give different input formats, but internally we are 
assured that the data has the same format."""
from __future__ import annotations

import warnings
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Iterable, Type

import networkx as nx
from numpy.typing import ArrayLike

from qgym.templates import Rewarder, Visualiser
from qgym.utils.input_validation import (
    check_adjacency_matrix,
    check_graph_is_valid_topology,
    check_instance,
    check_string,
)

if TYPE_CHECKING:
    Gridspecs = list[int | Iterable[int]] | tuple[int | Iterable[int]]


def parse_rewarder(rewarder: Rewarder | None, default: Type[Rewarder]) -> Rewarder:
    """Parse a `rewarder` given by the user.

    Args:
        rewarder: ``Rewarder`` to use for the environment. If ``None``, then a new
            instance of the `default` rewarder will be returned.
        default: Type of the desired default rewarder to used when no rewarder is given.

    Returns:
        A deepcopy of the given `rewarder` or a new instance of type `default` if
        `rewarder` is ``None``.
    """
    if rewarder is None:
        return default()
    check_instance(rewarder, "rewarder", Rewarder)
    return deepcopy(rewarder)


def parse_visualiser(
    render_mode: str | None, vis_type: Type[Visualiser], args: list[Any]
) -> None | Visualiser:
    """Parse a `Visualiser` by the render mode.

    Args:
        render_mode: If ``None`` return ``None``. Otherwise return a ``Visualiser`` of
            type `vis_type` with optional arguments given in `args`.
        vis_type: Type of ``Visualiser`` to make if `render_mode` is not ``None``.
        args: Additional argument to give to the init of the ``Visualiser`` if
            `vis_type` is not ``None``.

    Returns:
        If `render_mode` is ``None`` return ``None``. Otherwise return a ``Visualiser``
        of type `vis_type` with optional arguments given in `args`.
    """
    if render_mode is None:
        return None

    render_mode = check_string(render_mode, "render_mode", lower=True)
    return vis_type(render_mode, *args)


def parse_connection_graph(
    graph: nx.Graph | None = None,
    matrix: ArrayLike | None = None,
    grid_size: Gridspecs | None = None,
) -> nx.Graph:
    """Parse the user input (given in ``__init__``) to create a connection graph.

    Args:
        graph: ``networkx.Graph`` representation of the QPU topology.
        matrix: Adjacency matrix representation of the QPU topology.
        size: Size of the connection graph when the topology is a grid.

    Raises:
        ValueError: When `graph`, `matrix` and `grid_size` are all ``None``.

    Returns:
        Connection graph as a ``networkx.Graph``.
    """
    if graph is not None:
        if matrix is not None:
            warnings.warn("Both 'graph' and 'matrix' were given. Using 'graph'.")
        if grid_size is not None:
            warnings.warn("Both 'graph' and 'grid_size' were given. Using 'graph'.")

        check_graph_is_valid_topology(graph, "graph")

        # deepcopy the graphs for safety
        return deepcopy(graph)

    if matrix is not None:
        if grid_size is not None:
            warnings.warn("Both 'matrix' and 'grid_size' were given. Using 'matrix'.")
        matrix = check_adjacency_matrix(matrix)
        return nx.from_numpy_array(matrix)

    if grid_size is not None:
        # Generate connection grid graph
        return nx.grid_graph(grid_size)

    raise ValueError("No valid arguments for a connection graph were given")



"""This module contains the ``GateEncoder`` class which can be used for encoding gates
to integers and back.

Usage:
    >>> from qgym.utils import GateEncoder
    >>> encoder = GateEncoder().learn_gates(["x", "y", "z", "cnot", "measure"])
    >>> encoded_list = encoder.encode_gates(["x", "x", "measure", "z"])
    >>> print(encoded_list)
    [1, 1, 5, 3]
    >>> encoder.decode_gates(encoded_list)
    ['x', 'x', 'measure', 'z']
"""
from __future__ import annotations

import warnings
from typing import Any, Iterable, Mapping, Sequence, TypeVar, cast, overload

from qgym.custom_types import Gate

T = TypeVar("T")


class GateEncoder:
    """Learns a set of gates and creates a mapping to integers and back."""

    def __init__(self) -> None:
        """Initialize the ``GateEncoder``."""
        self.n_gates = 0
        self._encoding_dct: dict[str, int] = {}
        self._decoding_dct: dict[int, str] = {}
        self._longest_name = 0

    def learn_gates(self, gates: Iterable[str]) -> GateEncoder:
        """Learns the gates names from an ``Iterable`` and creates a mapping from unique
        gate names to integers and back.

        Args:
            gates: ``Iterable`` containing the names of the gates that should be
                learned. The ``Iterable`` can contain duplicate names.

        Returns:
            Self.
        """
        self.n_gates = 0
        for idx, gate_name in enumerate(gates, 1):
            if gate_name in self._encoding_dct:
                warnings.warn(f"'gates' contains multiple entries of {gate_name}")
            else:
                self._encoding_dct[gate_name] = idx
                self._decoding_dct[idx] = gate_name
                self._longest_name = max(self._longest_name, len(gate_name))
                self.n_gates += 1

        return self

    @overload
    def encode_gates(self, gates: str) -> int:
        ...

    @overload
    def encode_gates(self, gates: Mapping[str, T]) -> dict[int, T]:
        ...

    @overload
    def encode_gates(self, gates: Sequence[Gate]) -> list[Gate]:
        ...

    @overload
    def encode_gates(self, gates: set[str]) -> set[int]:
        ...

    @overload
    def encode_gates(self, gates: list[str] | tuple[str, ...]) -> list[int]:
        ...

    def encode_gates(
        self,
        gates: str
        | Mapping[str, Any]
        | Sequence[Gate]
        | set[str]
        | list[str]
        | tuple[str, ...],
    ) -> int | dict[int, Any] | list[Gate] | set[int] | list[int]:
        """Encode the gate names (of type ``str``) in `gates` to integers, based on the
        gates seen in ``learn_gates``.

        Args:
            gates: Gates to encode. The input type determines the return type.

        Raises:
            TypeError: When an unsupported type is given.

        Returns: Integer encoded version of gates. The output structure should resemble
            the input structure. So a ``Mapping`` will return a ``Dict``, while a single
            ``str`` will return an ``int``.
        """
        if isinstance(gates, str):
            encoded_str = self._encoding_dct[gates]
            return encoded_str

        if isinstance(gates, Mapping):
            return self._encode_mapping(gates)

        if isinstance(gates, Sequence) and isinstance(gates[0], Gate):
            # We assume that if the first element of gates is a Gate, then the whole
            # Sequence contains Gate objects.
            encoded_gates_list: list[Gate] = []
            for gate in cast(Sequence[Gate], gates):
                encoded_name = self._encoding_dct[gate.name]
                encoded_gates_list.append(Gate(encoded_name, gate.q1, gate.q2))
            return encoded_gates_list

        if isinstance(gates, set):
            encoded_names_set: set[int] = set()
            for gate_name in gates:
                gate_encoding = self._encoding_dct[gate_name]
                encoded_names_set.add(gate_encoding)
            return encoded_names_set

        if isinstance(gates, (list, tuple)):
            encoded_names_list: list[int] = []
            for gate_name in gates:
                gate_encoding = self._encoding_dct[gate_name]
                encoded_names_list.append(gate_encoding)
            return encoded_names_list

        raise TypeError(
            f"gates type must be str, Mapping or Sequence, got {type(gates)}."
        )

    def _encode_mapping(self, mapping: Mapping[str, Any]) -> dict[int, Any]:
        """Encode a mapping with gate names.

        Raises:
            ValueError: For unknown mappings.
        """
        encoded_dict: dict[int, Any] = {}
        gate_name: str
        for gate_name, item in mapping.items():
            gate_encoding = self._encoding_dct[gate_name]
            if isinstance(item, int):
                encoded_dict[gate_encoding] = item
            elif isinstance(item, Iterable):
                item_encoded = []
                for i in item:
                    item_encoded.append(self._encoding_dct[i])
                encoded_dict[gate_encoding] = item_encoded
            else:
                raise ValueError("Unknown mapping")
        return encoded_dict

    @overload
    def decode_gates(self, encoded_gates: int) -> str:
        ...

    @overload
    def decode_gates(self, encoded_gates: Mapping[int, Any]) -> dict[str, Any]:
        ...

    @overload
    def decode_gates(self, encoded_gates: Sequence[Gate]) -> list[Gate]:
        ...

    @overload
    def decode_gates(self, encoded_gates: set[int]) -> set[str]:
        ...

    @overload
    def decode_gates(self, encoded_gates: list[int] | tuple[int, ...]) -> list[str]:
        ...

    def decode_gates(
        self,
        encoded_gates: int
        | Mapping[int, Any]
        | Sequence[Gate]
        | set[int]
        | list[int]
        | tuple[int, ...],
    ) -> str | dict[str, Any] | list[Gate] | set[str] | list[str]:
        """Decode integer encoded gate names to the original gate names based on the
        gates seen in ``learn_gates``.

        Args:
            encoded_gates: Encoded gates that are to be decoded. The input type
            determines the return type.

        Raises:
            TypeError: When an unsupported type is given

        Returns:
            Decoded version of encoded_gates. The output structure should resemble the
            input structure. So a ``Mapping`` will return a ``Dict``, while a single
            ``int`` will return a ``str``.
        """
        if isinstance(encoded_gates, int):
            decoded_int = self._decoding_dct[encoded_gates]
            return decoded_int

        if isinstance(encoded_gates, Mapping):
            decoded_dict: dict[str, Any] = {}
            gate_int: int
            for gate_int in encoded_gates:
                gate_name = self._decoding_dct[gate_int]
                decoded_dict[gate_name] = encoded_gates[gate_int]
            return decoded_dict

        if isinstance(encoded_gates, Sequence) and isinstance(encoded_gates[0], Gate):
            # We assume that if the first element of encoded_gates is a Gate, then the
            # whole Sequence contains Gate objects.
            decoded_gate_list: list[Gate] = []
            for gate in cast(Sequence[Gate], encoded_gates):
                decoded_gate_name = self._decoding_dct[gate.name]
                decoded_gate_list.append(Gate(decoded_gate_name, gate.q1, gate.q2))
            return decoded_gate_list

        if isinstance(encoded_gates, set):
            decoded_name_set: set[str] = set()
            for gate_int in encoded_gates:
                decoded_gate = self._decoding_dct[gate_int]
                decoded_name_set.add(decoded_gate)
            return decoded_name_set

        if isinstance(encoded_gates, (list, tuple)):
            decoded_name_list: list[str] = []
            for gate_int in encoded_gates:
                decoded_gate = self._decoding_dct[gate_int]
                decoded_name_list.append(decoded_gate)
            return decoded_name_list

        raise TypeError(
            "encoded_gates must be int, Mapping or Sequence, got "
            f"{type(encoded_gates)}."
        )

    def __repr__(self) -> str:
        """Make a string representation without endline characters."""
        return f"{self.__class__.__name__}(encoding={self._encoding_dct})"



"""Module containing the environment, rewarders and visualizer for the initial mapping
problem of OpenQL.
"""
from qgym.envs.initial_mapping.initial_mapping import InitialMapping
from qgym.envs.initial_mapping.initial_mapping_rewarders import (
    BasicRewarder,
    EpisodeRewarder,
    SingleStepRewarder,
)
from qgym.envs.initial_mapping.initial_mapping_state import InitialMappingState

__all__ = [
    "InitialMapping",
    "InitialMappingState",
    "BasicRewarder",
    "EpisodeRewarder",
    "SingleStepRewarder",
]



"""This module contains types and type aliases used during visualisation."""
from typing import Tuple

import pygame

Color = Tuple[int, int, int]
Font = pygame.font.Font
Surface = pygame.surface.Surface

__all__ = ["Color", "Font", "Surface"]



"""This module contains the ``Discrete`` space, i.e., a range of integers.

A sample returns a randomly generated number within the bounds of the Discrete space.

Usage:
    >>> from qgym.spaces import Discrete
    >>> Discrete(3)
    Discrete(3)

"""
from __future__ import annotations

import gymnasium.spaces
from numpy.random import Generator


class Discrete(gymnasium.spaces.Discrete):
    """Discrete action/observation space for use in RL environments."""

    def __init__(
        self,
        n: int,
        start: int = 0,
        *,
        rng: Generator | None = None,
    ) -> None:
        """Initialize a Discrete space,  i.e., a range of integers.

        Args:
            n: The number of integer values in the Discrete space.
            start: The smallest element is the Discrete Space.
            rng: Random number generator to be used in this space, if ``None`` a new
                random number generator will be constructed.
        """
        super().__init__(n=n, start=start)
        self._np_random = rng  # this overrides the default behaviour of the gym space



"""This package contains general spaces that are to be used as action or observation
space in custom RL Environments. All spaces inherit from spaces of the ``gymnasium``
package.
"""
from qgym.spaces.box import Box
from qgym.spaces.dict import Dict
from qgym.spaces.discrete import Discrete
from qgym.spaces.multi_binary import MultiBinary
from qgym.spaces.multi_discrete import MultiDiscrete

__all__ = ["Box", "Dict", "Discrete", "MultiBinary", "MultiDiscrete"]



"""Generic abstract base class for visualisers of RL environments.

All visualisers should inherit from ``Visualiser``.
"""
from __future__ import annotations

from abc import abstractmethod
from typing import Any

import numpy as np
import pygame
from numpy.typing import NDArray

from qgym.utils.visualisation.typing import Color, Font, Surface


class RenderData:
    """Class containing usefull data for rendering like screen, font and colors."""

    __slots__ = ["screen", "font", "colors", "render_mode"]

    def __init__(
        self,
        screen: Surface,
        font: dict[str, Font],
        colors: dict[str, Color],
        render_mode: str,
    ) -> None:
        self.screen = screen
        self.font = font
        self.colors = colors
        self.render_mode = render_mode

    @property
    def screen_width(self) -> int:
        """Screen width. Alias for ``self.screen_dimensions[0]``."""
        return self.screen.get_width()

    @property
    def screen_height(self) -> int:
        """Screen height. Alias for ``self.screen_dimensions[1]``."""
        return self.screen.get_height()


class Visualiser:
    """Visualizer for the the current state of the problem."""

    # --- These attributes should be set in any subclass ---
    render_data: RenderData

    @abstractmethod
    def __init__(self, render_mode: str, *args: list[Any]):
        raise NotImplementedError

    @abstractmethod
    def render(self, state: Any) -> None | NDArray[np.int_]:
        """Render the current state using ``pygame``."""
        raise NotImplementedError

    def step(self, state: Any) -> None:
        """To be used during a step of the environment.

        Renders the display if `render_mode` is 'human', does nothing otherwise.

        Args:
            state: State to render if `render_mode` is 'human'.
        """
        if self.render_data.render_mode == "human":
            self.render(state)

    def _display(self) -> None | NDArray[np.int_]:
        """Display the current state using ``pygame``.

        The render function should call this method at the end.

        Raises:
            ValueError: When an invalid mode is provided.

        Returns:
            If 'human' mode returns a boolean value encoding whether the ``pygame``
            screen is open. In 'rgb_array' mode returns an RGB array encoding of the
            rendered image.
        """
        if self.render_data.render_mode == "human":
            pygame.event.pump()
            pygame.display.flip()
            return None

        if self.render_data.render_mode == "rgb_array":
            return np.transpose(
                np.array(pygame.surfarray.pixels3d(self.screen)), axes=(1, 0, 2)
            )

        msg = f"You provided an invalid mode '{self.render_data.render_mode}', the only"
        msg += " supported modes are 'human' and 'rgb_array'."
        raise ValueError(msg)

    def _start_screen(
        self, screen_name: str, render_mode: str, screen_dimensions: tuple[int, int]
    ) -> Surface:
        """Start a pygame screen in the given mode.

        Args:
            screen_name: Name of the screen.
            render_mode: The render mode to use. Choose from 'human' or 'rgb_array'.
            screen_dimension: Width and height of the screen.

        Raises:
            ValueError: When an invalid mode is provided.

        Returns:
            The initialized screen.
        """
        if not isinstance(render_mode, str):
            raise TypeError(
                f"'rendermode' of type {type(render_mode)} has no screen to start"
            )

        pygame.display.init()
        if render_mode == "human":
            screen = pygame.display.set_mode(screen_dimensions)
        elif render_mode == "rgb_array":
            screen = pygame.Surface(screen_dimensions)
        else:
            raise ValueError(
                f"You provided an invalid render mode '{render_mode}', the only "
                "supported modes are 'human' and 'rgb_array'."
            )
        pygame.display.set_caption(screen_name)
        return screen

    def close(self) -> None:
        """Close the screen used for rendering."""
        pygame.quit()

    @property
    def screen_width(self) -> int:
        """Screen width of  the main screen."""
        return self.render_data.screen_width

    @property
    def screen_height(self) -> int:
        """Screen height of the main screen."""
        return self.render_data.screen_height

    @property
    def is_open(self) -> bool:
        """Boolean value stating whether a ``pygame.screen`` is currently open."""
        return self.render_data.screen is not None

    @property
    def colors(self) -> dict[str, Color]:
        """Dict containing names color pairs."""
        return self.render_data.colors

    @property
    def screen(self) -> Surface:
        """Main screen to draw on."""
        return self.render_data.screen

    @property
    def font(self) -> dict[str, Font]:
        """Dict containing str Font pairs."""
        return self.render_data.font



"""Generic abstract base class for a RL rewarder.

All rewarders should inherit from ``Rewarder``.
"""
from __future__ import annotations

from abc import abstractmethod
from typing import Any


class Rewarder:
    """RL Rewarder, for computing rewards on a state."""

    _reward_range: tuple[float, float]

    @abstractmethod
    def compute_reward(self, *, old_state: Any, action: Any, new_state: Any) -> float:
        """Compute a reward, based on the old state, new state, and the given action.

        Args:
            old_state: State of the ``Environment`` before the current action.
            action: Action that has just been taken.
            new_state: Updated state of the ``Environment``.

        Returns:
            The reward for this action.
        """
        raise NotImplementedError

    @property
    def reward_range(self) -> tuple[float, float]:
        """Reward range of the rewarder. I.e., range that rewards can lie in."""
        return self._reward_range

    def __eq__(self, other: Any) -> bool:
        """Checks whether an object 'other' is equal to self.

        This check is performed by checking of the self and other are of exactly the
        same type. Afterwards, all slots (if any) are equal and if all attributes are
        equal.

        Returns:
            Boolean stating wether other is equal to self.
        """
        if type(self) is not type(other):
            return False

        if hasattr(self, "__slots__"):
            for attr in self.__slots__:
                if getattr(self, attr) != getattr(other, attr):
                    return False

        if hasattr(self, "__dict__") and self.__dict__ != other.__dict__:
            return False

        return True



r"""This module contains an environment for training an RL agent on the quantum
operation scheduling problem of OpenQL. The quantum operations scheduling problem is
aimed at finding the shortest possible schedules of operations defined by a **quantum
circuit**, whilst taking **hardware constraints** and **commutation rules** into
account.

Circuits:
    The quantum operations scheduling problem consist of scheduling the operations
    described by a quantum circuit. For the ``Scheduling`` environment in this module, a
    quantum circuit is defined as a list of gates, where a gate is a ``namedtuple`` with
    a 'name', 'q1' and 'q2'. For example, a c-not with control qubit 2 and target qubit
    5 would be ``Gate("cnot", 2, 5)`` and an X gate acting on qubit 4 would be
    ``Gate("x", 4, 4)``. In this representation, the circuit below can be created by:

    >>> from qgym.custom_types import Gate
    >>> circuit = [Gate("x", 1, 1), Gate("cnot", 0, 1), Gate("x", 0, 0), Gate("h",1,1)]

    .. code-block:: console

                  QUANTUM CIRCUIT
               ┌───┐   ┌───┐   ┌───┐
        |q1>───┤ Y ├───┤ X ├───┤ H ├──
               └───┘   └─┬─┘   └───┘
                         │     ┌───┐
        |q0>─────────────┴─────┤ X ├──
                               └───┘


Hardware specifications:
    Different operations defined by a quantum circuit have different operation times.
    These operation times are defined in the unit of machine cycles. For example, an X
    gate could take 2 machine cycles, whereas a measurement could take 15 cycles. These
    different operation times must be taken into account, because the hardware (the
    quantum computers) cannot perform multiple operations on the same qubit at the same
    time.

    Finding an optimal schedule with these restriction is already a difficult task.
    However, to complicate matters there can be more limitations defined by the
    hardware. To be more concrete, the ``Schdeuling`` environment can take two more
    types of limitations into account:

    #. Some gates must start at the same time, or must wait till the previous one is
       finished. This is typically the case for measurements.
    #. Some gates can't be executed in the same cycle. This for example happens often
       with X, Y and Z gates.

    The goal of the quantum operations scheduling problem is to find an optimal
    schedule, which takes these hardware limitations into account, as well as some
    optional commutation rules.

State space:
    The state space is described by a dictionary with the following structure:

    * `machine_properties`: MachineProperties object containing machine properties
      and limitations.
    * `utils`: ``SchedulingUtils`` dataclass with a random circuit generator,
      commutation rulebook and a gate encoder.
    * `gates`: Dictionary with gate names as keys and ``GateInfo`` dataclasses as
      values.
    * `steps_done`: Number of steps done since the last reset.
    * `cycle`: Current 'machine' cycle.
    * `busy`: Used internally for the hardware limitations.
    * `circuit_info`: ``CircuitInfo`` dataclass containing the encoded circuit and
      attributes used to update the state.

Observation space:
    Each element in the observation space is a dictionary with 4 entries:

    * `gate_names`: Gate names of the encoded circuit.
    * `acts_on`: q1 and q2 of each gate.
    * `dependencies`: Shows the first $n$ gates that must be scheduled before this
      gate.
    * `legal_actions`: List of legal actions. If the value at index $i$ determines if
      gate number $i$ can be scheduled or not.

Action Space:
    Performing a quantum operation takes a certain amount of time, which is measured in
    (machine) cycles. Therefore, this environment aims to produce a schedule in terms
    of cycles. To do this, the environment schedules the circuit from right to left,
    so cycle zero in the schedule is the last operation to be performed in that
    schedule. Based on this idea, a valid action is then a tuple of an integer
    $i\in\{0,1,\ldots,max\_gates\}$ and a binary value $j$. The integer value $i$
    schedules gate $i$ in the current cycle, and the binary value $j$ increment the
    cycle number.

    If the action is illegal (for example when we try to schedule a gate that has been
    scheduled already), then the environment will do nothing, and the rewarder should
    give a penalty for this.

Example 1:
    In this example we want to create an environment for a machine with the following
    properties:

    * The machine has two qubits.
    * The machine supports the X, Y, C-NOT and Measure gates.
    * Multiple Measure gates should start in the same cycle, or wait till the previous
      one is done.
    * The X and Y gates cannot be performed in the same cycle.

    The code block below shows how a ``Scheduling`` environment for such a machine, were
    the environment only allows for circuit with a maximum length of 5:

    .. code-block:: python

        from qgym.envs.scheduling import Scheduling, MachineProperties

        # Set up the hardware specifications
        hardware_spec = MachineProperties(n_qubits=2)
        hardware_spec.add_gates({"x": 2, "y": 2, "cnot": 4, "measure": 10})
        hardware_spec.add_same_start(["measure"])
        hardware_spec.add_not_in_same_cycle([("x", "y")])

        # Initialize the environment
        env = Scheduling(hardware_spec, max_gates=5, random_circuit_mode="workshop")


Example 2:
    This example shows how to add a specified commutation rulebook, which can contain
    costum commutation rules. For more information on implementing custom cummutation
    rules, see the documentation of ``CommutationRulebook``. This example uses the same
    machine properties as example 1. Costrum commutation rules can be added as shown in
    the code block below:

    .. code-block:: python

        from qgym.envs.scheduling import Scheduling, MachineProperties
        from qgym.envs.scheduling.rulebook import CommutationRulebook

        # Set up the hardware specifications
        hardware_spec = MachineProperties(n_qubits=2)
        hardware_spec.add_gates({"x": 2, "y": 2, "cnot": 4, "measure": 10})
        hardware_spec.add_same_start(["measure"])
        hardware_spec.add_not_in_same_cycle([("x", "y")])

        # Setup the rulebook
        rulebook = CommutationRulebook()

        # Initialize the environment
        env_com = Scheduling(
                            hardware_spec,
                            max_gates=5,
                            rulebook=rulebook,
                            random_circuit_mode="workshop",
                        )


"""
from __future__ import annotations

from copy import deepcopy
from typing import Any, Dict, Mapping, Union, cast

import numpy as np
from numpy.typing import NDArray

import qgym.spaces
from qgym.custom_types import Gate
from qgym.envs.scheduling.machine_properties import MachineProperties
from qgym.envs.scheduling.rulebook import CommutationRulebook
from qgym.envs.scheduling.scheduling_rewarders import BasicRewarder
from qgym.envs.scheduling.scheduling_state import SchedulingState
from qgym.envs.scheduling.scheduling_visualiser import SchedulingVisualiser
from qgym.templates import Environment, Rewarder
from qgym.utils.input_parsing import parse_rewarder, parse_visualiser
from qgym.utils.input_validation import check_instance, check_int, check_string


class Scheduling(
    Environment[Dict[str, Union[NDArray[np.int_], NDArray[np.int8]]], NDArray[np.int_]]
):
    """RL environment for the scheduling problem."""

    def __init__(
        self,
        machine_properties: Mapping[str, Any] | str | MachineProperties,
        *,
        max_gates: int = 200,
        dependency_depth: int = 1,
        random_circuit_mode: str = "default",
        rulebook: CommutationRulebook | None = None,
        rewarder: Rewarder | None = None,
        render_mode: str | None = None,
    ) -> None:
        """Initialize the action space, observation space, and initial states for the
        scheduling environment.

        Args:
            machine_properties: A ``MachineProperties`` object, a ``Mapping`` containing
                machine properties or a string with a filename for a file containing the
                machine properties.
            max_gates: Maximum number of gates allowed in a circuit. Defaults to 200.
            dependency_depth: Number of dependencies given in the observation.
                Determines the shape of the `dependencies` observation, which has the
                shape (dependency_depth, max_gates). Defaults to 1.
            random_circuit_mode: Mode for the random circuit generator. The mode can be
                'default' or 'workshop'. Defaults to 'default'.
            rulebook: ``CommutationRulebook`` describing the commutation rules. If
                ``None`` (default) is given, a default ``CommutationRulebook`` will be
                used. (See ``CommutationRulebook`` for more info on the default rules.)
            rewarder: Rewarder to use for the environment. If ``None`` (default), then a
                default ``BasicRewarder`` is used.
            render_mode: If 'human' open a ``pygame`` screen visualizing the step. If
                'rgb_array', return an RGB array encoding of the rendered frame on each
                render call.
        """
        self.metadata = {
            "render.modes": ["human", "rgb_array"],
            "random_circuit.mode": ["default", "workshop"],
        }

        machine_properties = self._parse_machine_properties(machine_properties)
        max_gates = check_int(max_gates, "max_gates", l_bound=1)
        dependency_depth = check_int(dependency_depth, "dependency_depth", l_bound=1)
        random_circuit_mode = self._parse_random_circuit_mode(random_circuit_mode)
        rulebook = self._parse_rulebook(rulebook)
        self._rewarder = parse_rewarder(rewarder, BasicRewarder)

        self._state = SchedulingState(
            machine_properties=machine_properties,
            max_gates=max_gates,
            dependency_depth=dependency_depth,
            random_circuit_mode=random_circuit_mode,
            rulebook=rulebook,
        )
        self.observation_space = self._state.create_observation_space()
        self.action_space = qgym.spaces.MultiDiscrete([max_gates, 2], rng=self.rng)

        self._visualiser = parse_visualiser(
            render_mode, SchedulingVisualiser, [self._state]
        )

    def reset(
        self,
        *,
        seed: int | None = None,
        options: Mapping[str, Any] | None = None,
    ) -> tuple[dict[str, NDArray[np.int_] | NDArray[np.int8]], dict[str, Any]]:
        """Reset the state, action space and load a new (random) initial state.

        To be used after an episode is finished.

        Args:
            seed: Seed for the random number generator, should only be provided
                (optionally) on the first reset call, i.e., before any learning is done.
            return_info: Whether to receive debugging info.
            options: Mapping with keyword arguments with additional options for the
                reset. Keywords can be found in the description of
                ``SchedulingState.reset()``
            _kwargs: Additional options to configure the reset.

        Returns:
            Initial observation and debugging info.
        """
        return super().reset(seed=seed, options=options)

    def get_circuit(self, mode: str = "human") -> list[Gate]:
        """Return the quantum circuit of this episode.

        Args:
            mode: Choose from be 'human' or 'encoded'. Defaults to 'human'.

        Raises:
            ValueError: If an unsupported mode is provided.

        Returns:
            Human or encoded quantum circuit.
        """
        mode = check_string(mode, "mode", lower=True)
        state = cast(SchedulingState, self._state)
        encoded_circuit = state.circuit_info.encoded
        if mode == "encoded":
            return deepcopy(encoded_circuit)

        if mode == "human":
            gate_encoder = state.utils.gate_encoder
            return gate_encoder.decode_gates(encoded_circuit)

        raise ValueError(f"mode must be 'human' or 'encoded', but was {mode}")

    @staticmethod
    def _parse_machine_properties(
        machine_properties: Mapping[str, Any] | str | MachineProperties
    ) -> MachineProperties:
        """
        Parse the machine_properties given by the user and return a
        ``MachineProperties`` object.

        Args:
            machine_properties: A ``MachineProperties`` object, a ``Mapping`` of machine
            or a string with a filename for a file containing the machine properties.

        Raises:
            TypeError: If the given type is not supported.

        Returns:
            ``MachineProperties`` object with the given machine properties.
        """
        if isinstance(machine_properties, str):
            return MachineProperties.from_file(machine_properties)
        if isinstance(machine_properties, MachineProperties):
            return deepcopy(machine_properties)
        if isinstance(machine_properties, Mapping):
            return MachineProperties.from_mapping(machine_properties)

        msg = f"{type(machine_properties)} is not a supported type for "
        msg += "'machine_properties'"
        raise TypeError(msg)

    @staticmethod
    def _parse_random_circuit_mode(random_circuit_mode: str) -> str:
        """Parse the `random_circuit_mode` given by the user.

        Args:
            random_circuit_mode: Mode for the random circuit generator. The mode can be
            'default' or 'workshop'.

        Raises:
            ValueError: If the given mode is not supported.

        Returns:
            Valid random circuit mode.
        """
        random_circuit_mode = check_string(
            random_circuit_mode, "random_circuit_mode", lower=True
        )
        if random_circuit_mode not in ["default", "workshop"]:
            msg = f"'{random_circuit_mode}' is not a supported random circuit mode"
            raise ValueError(msg)
        return random_circuit_mode

    @staticmethod
    def _parse_rulebook(rulebook: CommutationRulebook | None) -> CommutationRulebook:
        """Parse the `rulebook` given by the user.

        Args:
            rulebook: Rulebook describing the commutation rules. If ``None`` is given, a
                default ``CommutationRulebook`` will be used. (See
                ``CommutationRulebook`` for more info on the default rules.)

        Returns:
            ``CommutationRulebook``.
        """
        if rulebook is None:
            return CommutationRulebook()
        check_instance(rulebook, "rulebook", CommutationRulebook)
        return deepcopy(rulebook)



"""This module contains wrappers around some commonly used ``pygame`` functions."""
import numpy as np
import pygame
from numpy.typing import ArrayLike
from pygame import gfxdraw

from qgym.utils.visualisation.typing import Color, Font, Surface

# pylint: disable=invalid-name


def draw_point(screen: Surface, pos: ArrayLike, color: Color, r: int = 10) -> None:
    """Draw a point on the screen.

    Args:
        screen: Screen to add the point to.
        pos: ``ArrayLike`` containing the x and y coordinates of the point. Non integer
            values will be rounded down to the nearest integer.
        r: Radius of the point (in pixels). Defaults to 10.
    """
    pos_x, pos_y = np.asarray(pos, dtype=int)
    gfxdraw.aacircle(screen, pos_x, pos_y, r, color)
    gfxdraw.filled_circle(screen, pos_x, pos_y, r, color)


def draw_wide_line(
    screen: Surface,
    color: Color,
    point1: ArrayLike,
    point2: ArrayLike,
    *,
    width: int = 2
) -> None:
    """Draw a wide line on the screen.

    Args:
        screen: Screen to draw the line on.
        color: Color of the line.
        point1: Coordinates of the starting point of the line.
        point2: Coordinates of the end point of the line.
        width: Width of the line. Defaults to 2.
    """
    # distance between the points
    p1 = np.asarray(point1)
    p2 = np.asarray(point2)
    distance = np.linalg.norm(p2 - p1)

    # scaled perpendicular vector (vector from p1 & p2 to the polygon's points)
    sp = np.array([p1[1] - p2[1], p2[0] - p1[0]]) * 0.5 * width / distance

    # points
    points = (p1 - sp, p1 + sp, p2 + sp, p2 - sp)

    # draw the polygon
    pygame.gfxdraw.aapolygon(screen, points, color)  # type: ignore[arg-type]
    pygame.gfxdraw.filled_polygon(screen, points, color)  # type: ignore[arg-type]


def write_text(
    screen: Surface, font: Font, text: str, pos: ArrayLike, color: Color
) -> None:
    """Write text, centered around the given point.

    Args:
        screen: Screen to write the text on.
        font: Font to use.
        text: Text to write.
        pos: x and y coordinates of the center of the text.
        color: Color of the text.
    """
    pos_x, pos_y = np.asarray(pos, dtype=int)
    pygame_text = font.render(text, True, color)
    text_position = pygame_text.get_rect(center=(pos_x, pos_y))
    screen.blit(pygame_text, text_position)


def shade_rect(
    screen: Surface, size: ArrayLike, pos: ArrayLike, color: Color, alpha: int
) -> None:
    """Shade a rectangular area.

    Args:
        screen: Screen to draw the shade on.
        size: Width and height of the rectangular area.
        pos: The x,y-coordinates of the lower left corner of the rectangle.
        color: Color of the shaded area.
        alpha: Alpha of the shade.
    """
    rect_width, rect_height = np.asarray(size, dtype=int)
    pos_x, pos_y = np.asarray(pos, dtype=int)

    surf = pygame.Surface((rect_width, rect_height))
    surf.set_alpha(alpha)
    surf.fill(color)
    screen.blit(surf, (pos_x, pos_y))



"""Generic abstract base class for RL environments.

All environments should inherit from ``Environment``.
"""
from __future__ import annotations

from abc import abstractmethod
from copy import deepcopy
from typing import Any, Mapping

import gymnasium
import numpy as np
from gymnasium import Space
from numpy.random import Generator, default_rng
from numpy.typing import NDArray

from qgym.templates.rewarder import Rewarder
from qgym.templates.state import ActionT, ObservationT, State
from qgym.templates.visualiser import Visualiser


class Environment(gymnasium.Env[ObservationT, ActionT]):
    """RL Environment containing the current state of the problem.

    Each subclass should set at least the following attributes:
    """

    # --- These attributes should be set in any subclass ---
    action_space: Space[Any]
    """The action space of this environment."""
    observation_space: Space[Any]
    """The observation space of this environment."""
    metadata: dict[str, Any]
    """Additional metadata of this environment."""
    _state: State[ObservationT, ActionT]
    """The state space of this environment."""
    _rewarder: Rewarder
    """The rewarder of this environment."""
    _visualiser: Visualiser | None = None
    """The visualiser of this environment."""

    # --- Other attributes ---
    _rng: Generator | None = None

    def step(
        self, action: ActionT
    ) -> tuple[ObservationT, float, bool, bool, dict[Any, Any]]:
        """Update the state based on the input action.

        Return observation, reward, done-indicator, terminated-indicator and debugging
        info based on the updated state.

        Args:
            action: Action to be performed.

        Returns:
            A tuple containing five entries

            1. The updated state;
            2. Reward for the given action;
            3. Boolean value stating whether the new state is a final state (i.e., if
               we are done);
            4. Boolean value stating whether the episode is truncated. Currently always
               returns ``False``.
            5. Additional (debugging) information.
        """
        old_state = deepcopy(self._state)
        self._state.update_state(action)
        if self._visualiser is not None:
            self._visualiser.step(self._state)

        return (
            self._state.obtain_observation(),
            self._compute_reward(old_state, action),
            self._state.is_done(),
            False,
            self._state.obtain_info(),
        )

    @abstractmethod
    def reset(
        self, *, seed: int | None = None, options: Mapping[str, Any] | None = None
    ) -> tuple[ObservationT, dict[str, Any]]:
        """Reset the environment and load a new random initial state.

        To be used after an episode is finished. Optionally, one can provide additional
        options to configure the reset.

        Args:
            seed: Seed for the random number generator, should only be provided
                (optionally) on the first reset call, i.e., before any learning is done.
            options: Dictionary containing keyword-argument pairs to configure the
                reset.

        Returns:
            Initial observation and a dictionary containing debugging information.
        """
        super().reset(seed=seed)
        options = {} if options is None else options
        self._state.reset(seed=seed, **options)
        self.render()
        return self._state.obtain_observation(), self._state.obtain_info()

    def render(self) -> None | NDArray[np.int_]:  # type: ignore[override]
        """Render the current state using pygame.

        Returns:
            Result of rendering.
        """
        if self._visualiser is not None:
            return self._visualiser.render(self._state)
        return None

    def close(self) -> None:
        """Close the screen used for rendering."""
        if self._visualiser is not None:
            self._visualiser.close()
        self._visualiser = None

    @property
    def rewarder(self) -> Rewarder:
        """Return the rewarder that is set for this environment.

        Used to compute rewards after each step.
        """
        return self._rewarder

    @rewarder.setter
    def rewarder(self, rewarder: Rewarder) -> None:
        self._rewarder = rewarder
        self.reward_range = rewarder.reward_range

    @property
    def rng(self) -> Generator:
        """Return the random number generator of this environment.

        If none is set yet, this will generate a new one using
        ``numpy.random.default_rng``.
        """
        if self._rng is None:
            self._rng = default_rng()
        return self._rng

    @rng.setter
    def rng(self, rng: Generator) -> None:
        self._rng = rng

    def __del__(self) -> None:
        self.close()

    def _compute_reward(
        self,
        old_state: State[ObservationT, ActionT],
        action: ActionT,
        *args: Any,
        **kwargs: Any,
    ) -> float:
        """Ask the ``Rewarder`` to compute a reward, based on the given old state, the
        given action and the updated state.

        Args:
            old_state: The state of the ``Environment`` before the action was taken.
            action: Action that was taken.
            args: Optional arguments for the ``Rewarder``.
            kwargs: Optional keyword-arguments for the ``Rewarder``.
        """
        return self._rewarder.compute_reward(
            *args, old_state=old_state, action=action, new_state=self._state, **kwargs
        )



"""Generic utils for the Reinforcement Learning QGym."""
from qgym.utils.gate_encoder import GateEncoder
from qgym.utils.random_circuit_generator import RandomCircuitGenerator

__all__ = ["GateEncoder", "RandomCircuitGenerator"]



r"""This module contains an environment for training an RL agent on the routing problem
of OpenQL. The routing problem is aimed at enabling to execute the quantum circuit
by putting those physical qubits into connection that have an interaction in the quantum
circuit. This problem arises when there are mismatches between the interaction graph and
the QPU-topology in the initial mapping. The quantum circuit is represented as an
**interaction graph**, where each node represent a qubit and each edge represent an
interaction between two qubits as defined by the circuit (See the example below). The
QPU structure is called the **connection graph**. In the connection graph each node
represents a physical qubit and each edge represent a connection between two qubits in
the QPU.


.. code-block:: console

              QUANTUM CIRCUIT                        INTERACTION GRAPH
           ┌───┐               ┌───┐
    |q3>───┤ R ├───┬───────────┤ M ╞══                 q1 ────── q2
           └───┘   │           └───┘                            ╱
           ┌───┐ ┌─┴─┐         ┌───┐                           ╱
    |q2>───┤ R ├─┤ X ├───┬─────┤ M ╞══                        ╱
           └───┘ └───┘   │     └───┘                         ╱
           ┌───┐       ┌─┴─┐   ┌───┐                        ╱
    |q1>───┤ R ├───┬───┤ X ├───┤ M ╞══                     ╱
           └───┘   │   └───┘   └───┘                      ╱
           ┌───┐ ┌─┴─┐         ┌───┐                     ╱
    |q0>───┤ R ├─┤ X ├─────────┤ M ╞══                q3 ─────── q4
           └───┘ └───┘         └───┘


A SWAP-gate changes the mapping from logical qubits to physical qubits at a certain
point in the circuit, and thereby allows to solve mismatchings from the initial mapping.
The goal is to place SWAP-gates in the quantum circuit to fix the mismatches. The least
amount of SWAP-gates is preferred. In more advanced setups, also different factors can
be taken into account, like the fidelity of connections in the QPU.


State Space:
    The state space is described by a ``RoutingState`` with the following
    attributes:

    * `steps_done`: Number of steps done since the last reset.
    * `num_nodes`: Number of *physical* qubits.
    * `connection_graph`: A networkx representation of the connection graph.
    * `mapping`: Array of which the index represents a physical qubit, and the value a
      virtual qubit. This is updated after each swap.
    * `max_interaction_gates`: Maximum amount of gates allowed in the interaction 
      circuit, when a new interaction circuit is (randomly) generated.
    * `interaction_circuit`: An array of 2-tuples of integers, where every tuple
      represents a, not specified, gate acting on the two qubits labeled by the
      integers in the tuples.
    * `position`: The position in the original connection circuit.
    * `max_observation_reach`:  Caps the maximum amount of gates the agent can see ahead
      when making an observation.
    * `observe_legal_surpasses`: If ``True`` a list called `boolean_flags` will be
      added to the observation space. The list `boolean_flags` has length
      `observation_reach` and containing Boolean values indicating whether the gates
      ahead can be executed.
    * `observe_connection_graph`: If ``True``, the connection_graph will be incorporated
      in the observation_space.
    * `swap_gates_inserted`: A list of 3-tuples of integers, to register which gates to
      insert and where. Every tuple (g, q1, q2) represents the insertion of a SWAP-gate
      acting on logical qubits q1 and q2 before gate g in the interaction_circuit.

Observation Space:
    The observation space is a ``qgym.spaces.Dict`` with 2-4 entries:

    * `interaction_gates_ahead`: Array with Boolean values for the upcoming connection
      gates in the quantum circuit.
    * `mapping`: The current state of the mapping.
    * (Optional)`connection_graph`: Adjacency matrix of the connection graph.
    * (Optional)`is_legal_surpass_booleans`: Array with boolean values stating whether a
      connection gate can be surpassed with the current mapping.

Action Space:
    A valid action is a tuple of integers  $(i,j,k)$. The integer $i$ indicates whether
    the agent wants to surpass the current gate and move on to the next gate. If $i$ is 
    0, then a SWAP gate is inserted at the current position between qubits $j$ and $k$.
    Only legal actions will be executed, an action is legal when:

    #. $i=1$ and the next gate can be executed.
    #. $i=0$ and physical qubit $j$ does not equal physical qubit $k$.
    #. $i=0$ and physical qubits $j$ and $k$ have a connection between them in the
       connection graph.


# TODO: create Examples


"""
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterable, Mapping

import networkx as nx
import numpy as np
from numpy.typing import ArrayLike, NDArray

import qgym.spaces
from qgym.envs.routing.routing_rewarders import BasicRewarder
from qgym.envs.routing.routing_state import RoutingState
from qgym.envs.routing.routing_visualiser import RoutingVisualiser
from qgym.templates import Environment, Rewarder
from qgym.utils.input_parsing import (
    parse_connection_graph,
    parse_rewarder,
    parse_visualiser,
)
from qgym.utils.input_validation import check_bool, check_int

if TYPE_CHECKING:
    Gridspecs = list[int | Iterable[int]] | tuple[int | Iterable[int]]


class Routing(Environment[Dict[str, NDArray[np.int_]], NDArray[np.int_]]):
    """RL environment for the routing problem of OpenQL."""

    def __init__(
        self,
        max_interaction_gates: int = 10,
        max_observation_reach: int = 5,
        observe_legal_surpasses: bool = True,
        observe_connection_graph: bool = True,
        *,
        connection_graph: nx.Graph | None = None,
        connection_graph_matrix: ArrayLike | None = None,
        connection_grid_size: Gridspecs | None = None,
        rewarder: Rewarder | None = None,
        render_mode: str | None = None,
    ) -> None:
        """Initialize the action space, observation space, and initial states.

        The supported render modes of this environment are "human" and "rgb_array".

        Args:
            max_interaction_gates: Sets the maximum amount of gates in the
                `interaction_circuit`, when a new `interaction_circuit` is generated.
            max_observation_reach: Sets a cap on the maximum amount of gates the agent
                can see ahead when making an observation. When bigger that
                `max_interaction_gates` the agent will always see all gates ahead in an
                observation
            observe_legal_surpasses: If ``True`` a boolean array of length
                `observation_reach` indicating whether the gates ahead can be executed,
                will be added to the `observation_space`.
            observe_connection_graph: If ``True``, the connection_graph will be
                incorporated in the observation_space. Reason to set it ``False`` is:
                QPU-topology practically doesn't change a lot for one machine, hence an
                agent is typically trained for just one QPU-topology which can be
                learned implicitly by rewards and/or the booleans if they are shown,
                depending on the other flag above.
            connection_graph: ``networkx`` graph representation of the QPU topology.
                Each node represents a physical qubit and each node represents a
                connection in the QPU topology.
            connection_graph_matrix: Adjacency matrix representation of the QPU
                topology.
            connection_grid_size: Size of the connection graph when the connection graph
                has a grid topology. For more information on the allowed values and
                types, see ``networkx`` `grid_graph`_ documentation.
            rewarder: Rewarder to use for the environment. Must inherit from
                ``qgym.Rewarder``. If ``None`` (default), then ``BasicRewarder`` is
                used.
            render_mode: If 'human' open a ``pygame`` screen visualizing the step. If
                'rgb_array', return an RGB array encoding of the rendered frame on each
                render call.

        .. _grid_graph: https://networkx.org/documentation/stable/reference/generated/
            networkx.generators.lattice.grid_graph.html#grid-graph
        """
        # Check user input and parse it to a uniform format
        connection_graph = parse_connection_graph(
            connection_graph, connection_graph_matrix, connection_grid_size
        )

        max_interaction_gates = check_int(
            max_interaction_gates, "max_interaction_gates", l_bound=0
        )
        max_observation_reach = check_int(
            max_observation_reach,
            "max_observation_reach",
            l_bound=0,
            u_bound=max_interaction_gates,
        )
        observe_legal_surpasses = check_bool(
            observe_legal_surpasses, "observe_legal_surpasses", safe=False
        )
        observe_connection_graph = check_bool(
            observe_connection_graph, "observe_connection_graph", safe=False
        )

        # Define internal attributes
        self._rewarder = parse_rewarder(rewarder, BasicRewarder)

        self._state = RoutingState(
            max_interaction_gates=max_interaction_gates,
            max_observation_reach=max_observation_reach,
            connection_graph=connection_graph,
            observe_legal_surpasses=observe_legal_surpasses,
            observe_connection_graph=observe_connection_graph,
        )
        self.observation_space = self._state.create_observation_space()

        # Define attributes defined in parent class
        self.action_space = qgym.spaces.MultiDiscrete(
            nvec=[
                2,
                connection_graph.number_of_nodes(),
                connection_graph.number_of_nodes(),
            ],
            rng=self.rng,
        )

        self.metadata = {"render.modes": ["human", "rgb_array"]}
        self._visualiser = parse_visualiser(
            render_mode, RoutingVisualiser, [connection_graph]
        )

    def reset(
        self,
        *,
        seed: int | None = None,
        options: Mapping[str, Any] | None = None,
    ) -> tuple[dict[str, NDArray[np.int_]], dict[str, Any]]:
        """Reset the state and set/create a new interaction circuit.

        To be used after an episode is finished.

        Args:
            seed: Seed for the random number generator, should only be provided
                (optionally) on the first reset call i.e., before any learning is done.
            options: Mapping with keyword arguments with additional options for the
                reset. Keywords can be found in the description of
                ``RoutingState.reset()``

        Returns:
            Initial observation and debugging info.
        """
        # call super method for dealing with the general stuff
        return super().reset(seed=seed, options=options)



"""This package contains templates used from buidling custom environments.

All ``Environment``s should inherit from ``Environment`` and should contain a rewarder,
state and visualiser which inherit from the base classes ``Rewarder``, ``State`` and
``Visualiser`` respectively.
"""
from qgym.templates.environment import Environment
from qgym.templates.rewarder import Rewarder
from qgym.templates.state import State
from qgym.templates.visualiser import Visualiser

__all__ = ["Environment", "Rewarder", "State", "Visualiser"]



"""This module contains the ``MultiDiscrete`` space, i.e., multiple discrete intervals.
A sample returns one item from each interval.

Usage:
    >>> from qgym.spaces import MultiDiscrete
    >>> MultiDiscrete(nvec=[2,3,4])
    MultiDiscrete([2 3 4])

"""
from __future__ import annotations

from typing import Any, Type

import gymnasium.spaces
import numpy as np
from numpy.random import Generator
from numpy.typing import ArrayLike


class MultiDiscrete(gymnasium.spaces.MultiDiscrete):
    """Multi-discrete action/observation space for use in RL environments."""

    def __init__(
        self,
        nvec: ArrayLike,
        dtype: str | Type[np.integer[Any]] = np.int_,
        *,
        rng: Generator | None = None,
    ) -> None:
        """Initialize a multi-discrete space, i.e., multiple discrete intervals of given
        sizes.

        Args:
            nvec: Vector containing the upper bound of each discrete interval. The lower
                bound is always set to 0.
            dtype: Type of the values in each interval (default np.int64).
            rng: Random number generator to be used in this space, if ``None`` a new
                random number generator will be constructed.
        """
        super().__init__(nvec=np.asarray(nvec), dtype=dtype)
        self._np_random = rng  # this overrides the default behaviour of the gym space
